# Django settings for stuph project.

try:
    from stuphos import getConfig, getSection
    from stuphos.etc import isYesValue
except ImportError as e:
    print(f'[webserver.settings] {e}') # ok..
    DEBUG = True

djangoServiceConf = getSection('DjangoService')

def srvcConfig(name):
    return [_f for _f in (djangoServiceConf.get(name) or '').split('\n') if _f]

SHOW_DEBUG_PAGE = djangoServiceConf.get('show-debug-page')
DEBUG = SHOW_DEBUG_PAGE != 'admin' and isYesValue(SHOW_DEBUG_PAGE or 'no')
TEMPLATE_DEBUG = DEBUG

ADMINS = (
    # ('Your Name', 'your_email@domain.com'),
)

MANAGERS = ADMINS

from os.path import dirname, join as joinpath, realpath

WEB_PACKAGE_PATH = dirname(dirname(dirname(realpath(__file__)))) # stuphos/
PACKAGES_PATH = dirname(WEB_PACKAGE_PATH)

##    DATABASE_ENGINE = 'sqlite3'    # 'postgresql_psycopg2', 'postgresql', 'mysql', 'sqlite3' or 'oracle'.
##    DATABASE_NAME = DB_PATH        # Or path to database file if using sqlite3.
##    DATABASE_USER = ''             # Not used with sqlite3.
##    DATABASE_PASSWORD = ''         # Not used with sqlite3.
##    DATABASE_HOST = ''             # Set to empty string for localhost. Not used with sqlite3.
##    DATABASE_PORT = ''             # Set to empty string for default. Not used with sqlite3.
##
##    DATABASES = {
##        'default': {
##            'NAME': DB_PATH,
##            'ENGINE': 'django.db.backends.sqlite3',
##            'USER': '',
##            'PASSWORD': ''
##        }
##    }

##    DATABASE_ENGINE = 'postgresql_psycopg2'    # 'postgresql_psycopg2', 'postgresql', 'mysql', 'sqlite3' or 'oracle'.
##    DATABASE_NAME = 'stuph'        # Or path to database file if using sqlite3.
##    DATABASE_USER = 'stuph'             # Not used with sqlite3.
##    DATABASE_PASSWORD = 'password'         # Not used with sqlite3.
##    DATABASE_HOST = ''             # Set to empty string for localhost. Not used with sqlite3.
##    DATABASE_PORT = ''             # Set to empty string for default. Not used with sqlite3.

#from person.services.data.pg import getDjangoDatabaseConfig
from stuphos.db.conf import initDjangoDatabaseConfig
#import pdb; pdb.set_trace()

#mud.reloadConfigFile('support/.mud-config.cfg')

database = djangoServiceConf.get('database') or 'primary'

DATABASES = {}

try: DATABASES['default'] = initDjangoDatabaseConfig(database)
except KeyError as e:
    print(f'[webserver.settings] {database}: {e}')

# In case default gets overwritten, we always have access to the conf here.
try: DATABASES['stuphos'] = DATABASES['default']
except KeyError: pass

# XXX hardcoded namespace:
try: DATABASES['stuphosweb'] = initDjangoDatabaseConfig('sqlite')
except KeyError: pass


# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# If running in a Windows environment this must be set to the same as your
# system time zone.
TIME_ZONE = 'America/Chicago'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# Absolute path to the directory that holds media.
# Example: "/home/media/media.lawrence.com/"
MEDIA_ROOT = ''

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash if there is a path component (optional in other cases).
# Examples: "http://media.lawrence.com", "http://example.com/media/"
MEDIA_URL = ''

# URL prefix for admin media -- CSS, JavaScript and images. Make sure to use a
# trailing slash.
# Examples: "http://foo.com/media/", "/media/".
ADMIN_MEDIA_PREFIX = '/admin/media/'

# Make this unique, and don't share it with anybody.
SECRET_KEY_PATH = djangoServiceConf.get('secret-key-path', 'etc/django.secret.key')
try: SECRET_KEY = open(SECRET_KEY_PATH).read().strip()
except IOError as e:
    print('ERROR: Using default django site secret key.', e)
    SECRET_KEY = 'lla4r9dp9g8x)q4kuod-biqz0+=&-7&p(&p5$(1a9igwcr%+0%'


CMS_PROJECT_NAME = 'cms' # Name of directory in packages/web/
CMS_PROJECT_PATH = WEB_PACKAGE_PATH
CMS_PROJECT_PATH = joinpath(CMS_PROJECT_PATH, CMS_PROJECT_NAME)

CMS_MEDIA_ROOT = joinpath(CMS_PROJECT_PATH, 'media')
CMS_MEDIA_URL = '/cms/media/'

gettext = (lambda text: text)
CMS_TEMPLATES = (
    # ('page_base.html', gettext('StuphMUD Site Default')),
    # ('cms_base.html',  gettext('StuphCMS Base Page')),
)

LOGIN_URL = '/accounts/login'
LOGIN_REDIRECT_URL = '/' # menu.  "after login where to redirect (?)"

STATIC_ROOT = 'web/cms/media' # in lib: directory with cms -> package static dir
STATIC_URL = '/media/'

SEKIZAI_IGNORE_VALIDATION = True
SEKIZAI_VARNAME = 'my_sekizai_VarName'


DEFAULT_VMTEMPLATES = 'vm/ph/interpreter/mental/library/templates'

# Relative to stuphos package python path.  Section is not DjangoService.
VMTEMPLATES = getConfig('vm-templates', 'AgentSystem') or DEFAULT_VMTEMPLATES

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [joinpath(WEB_PACKAGE_PATH, 'server/templates'), # stuphos/server/templates
                 joinpath(PACKAGES_PATH, 'implementations/person/services/web/templates'),
                 joinpath(PACKAGES_PATH, VMTEMPLATES)] +

                srvcConfig('templates'),

        'APP_DIRS': False, # True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],

            'loaders': ['django.template.loaders.filesystem.Loader',
                        'django.template.loaders.app_directories.Loader',
                        ] + (['ph.interpreter.mental.library.loader.TemplateLoader']
                             if isYesValue(getConfig('virtual-templates', 'AgentSystem'))
                             else []),

            'builtins': [
                'phsite.network.templatetags.style',
                # 'phsite.network.templatetags.tools',
                # 'phsite.network.templatetags.world_lib',
            ]
        },
    },

    # {
    #     'BACKEND': 'person.features.girl.LibraryTemplates'
    # }
]

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
##    'django.template.loaders.filesystem.load_template_source',
##    'django.template.loaders.app_directories.load_template_source',
##    'web.stuph.embedded.entities.load_template_source',
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
#    'web.stuph.embedded.entities',
#     'django.template.loaders.eggs.load_template_source',
)

# MIDDLEWARE_CLASSES
MIDDLEWARE = (
    ##    'django.middleware.common.CommonMiddleware',
    ##    'django.contrib.sessions.middleware.SessionMiddleware',
    ##    'django.contrib.auth.middleware.AuthenticationMiddleware',

#        'django.middleware.cache.UpdateCacheMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',

    # Does this work? -- It requires apps to be configured first: see embedded.service.py
    'django.contrib.auth.middleware.AuthenticationMiddleware',

    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.common.CommonMiddleware',
#        'django.middleware.doc.XViewMiddleware',
        # 'django.middleware.csrf.CsrfViewMiddleware',
#        'cms.middleware.multilingual.MultilingualURLMiddleware',
#        'cms.middleware.page.CurrentPageMiddleware',
#        'cms.middleware.user.CurrentUserMiddleware',
#        'cms.middleware.toolbar.ToolbarMiddleware',
        # 'cms.middleware.media.PlaceholderMediaMiddleware',
        'django.middleware.cache.FetchFromCacheMiddleware',
)

if configurationTruth.DjangoService.log_all_requests:
    MIDDLEWARE += ('stuphos.webserver.LoggingMiddleware',)
if configurationTruth.DjangoService.multitenancy:
    MIDDLEWARE += ('stuphos.webserver.MultiTenancy',)
# if configurationTruth.DjangoService.csrf_exempt_middleware:
#     MIDDLEWARE += ('stuphos.webserver.CsrfExemptViewMiddleware',)


MULTITENANCY = {
}


_multitenancy = getSection('DjangoService:Multitenancy')
if _multitenancy is not None:
    for opt in _multitenancy.options():
        MULTITENANCY[opt] = _multitenancy[opt]


# XXX in application config.
# ROOT_URLCONF = 'stuphos.webserver.project.urls'

#import menus
#CMS_MENUS_PATH = dirname(menus.__file__)

TEMPLATE_DIRS = (
    # Put strings here, like "/home/html/django_templates" or "C:/www/django/templates".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
    joinpath(CMS_PROJECT_PATH, 'templates'),
    joinpath(CMS_PROJECT_PATH, 'site_templates'),
    # joinpath(CMS_MENUS_PATH, 'templates'),
)

TEMPLATE_CONTEXT_PROCESSORS = (
    'django.contrib.auth.context_processors.auth',
#    'django.core.context_processors.auth',
#    'django.core.context_processors.i18n',
#    'django.core.context_processors.request',
#    'django.core.context_processors.media',
#    'cms.context_processors.media',
#    'sekizai.context_processors.sekizai',
)

AUTHENTICATION_BACKENDS = ['django.contrib.auth.backends.ModelBackend',
                            # XXX
                            # 'phsite.network.accounts.KeyfileBackend'
                            ]

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
#    'django.contrib.databrowse',
    # 'django.contrib.admindocs',
    # 'django.contrib.admin',

    'django.contrib.messages',
#        'cms',
#        'cms.plugins.text',
        # 'cms.plugins.picture',
#        'cms.plugins.link',
#        'cms.plugins.file',
#        'cms.plugins.snippet',
#        'cms.plugins.googlemap',
#        'mptt',
        # 'publisher',
#        'menus',

#    'sekizai',

    'phsite.network',
    # 'taskqueues',
) + tuple(srvcConfig('webapps'))

# APPEND_SLASH = True

hostsConf = srvcConfig('hosts')

ALLOWED_HOSTS = (
    #u'*',
    'localhost',
) + tuple(hostsConf)


CSRF_TRUSTED_ORIGINS = \
    [
        # 'https://localhost:2180'
    ] + srvcConfig('csrf-trusted-origins')


# debugOn()
uploadMaxMemorySize = djangoServiceConf.get('upload-max-memory-size')
if uploadMaxMemorySize:
    DATA_UPLOAD_MAX_MEMORY_SIZE = int(uploadMaxMemorySize)


# Default settings for 'identities/{name}'.structure.preferences
defaultGlobalConfig_identities = dict()

CSRF_EXEMPT_PATHS = [
    'script/execute',
]



# mud.lang.structure
# from op.runtime.structural import document as _document
from . import structural as _document
getContextVariable = _document.Context.__getitem__

from stuphos.kernel import Machine as VM

def getContextEnvironment(name, **kwd):
	loader = getContextVariable('loader')

	try: default = kwd['default']
	except KeyError:
		return loader.environ[name]
	else:
		return loader.environ.get(name, default)

from stuphos.structure import Factory, SystemFactory, HTMLFactory

def load(source, classes, defaultName, **kwd):
	source = source.replace('\r', '') # scrub from any source.
	# source = 'Westmetal Configuration::\n\n' + source
	source = 'WMC []\n\n' + source

	# debugOn()
	try: return _document.loadStructuredMessageFromClasses \
				 (source, classes, raw = True,
				  default = defaultName, **kwd)

	except VM.Yield as e:
		# Don't propogate continuations because the document load will be
		# replaced by the yielding factory load, which might be confusing.
		raise RuntimeError('Incompatible continuation stack!') from e


def getFactories():
	yield 'stuph', Factory

	from stuphos import getConfig
	from stuphos.etc import isYesValue

	if isYesValue(getConfig('converge-spatial')):
		from spatial.spherical import structure as spatial
		yield 'world', spatial.Factory

def document(source, **kwd):
	# A raw load skips installation of core factory classes, so we
	# can create a controlled, discriminatory environment.
	# Auto WMC
	#import pdb; pdb.set_trace()
	return load(source, dict(getFactories()), 'stuph', **kwd)

def html(source, **kwd):
	return load(source, dict(getFactories(), html = HTMLFactory), 'html', **kwd)

def system(source, **kwd):
	return load(source, dict(system = SystemFactory), 'system', **kwd)

def resource(base, path, *names):
	base = io.path(base).folder(*path)
	return access(document(base.read()), *names)



# Auto-Slay Senario
# Copyright 2010 Clint Banis.  All rights reserved.
#
from . import Senario, QuestControl, Kitty
from . import MobilePrototype, ItemPrototype, RealRoom
from . import dumpYamlWithIndent

class AutoSlay(Senario):
    def __str__(self):
        def description():
            yield 'Name: %s' % getattr(self, 'name', '--')
            for (name, attribute) in [('Class'       , 'class'       ),
                                      ('Type'        , 'quest-type'  ),
                                      ('Field'       , 'field'       ),
                                      ('Min-Level'   , 'min-level'   ),
                                      ('Max-Level'   , 'max-level'   ),
                                      ('Remort-Level', 'remort-level'),
                                      ('Time-Limit'  , 'time-limit'  ),
                                      ('Availability', 'frequency'   ),
                                      ('Cost'        , 'cost'        ),
                                      ('Requirements', 'requirements'),
                                      ('Rewards'     , 'reward'      )]:

                yield '  %s: %s' % (name, self.get(attribute, '--'))

            try: opponents = self.opponents
            except AttributeError: pass
            else:
                yield '  Opponents:'
                for opp in opponents:
                    yield '    %s' % dumpYamlWithIndent(opp, indent = 4)

            try: rewards = self.rewards
            except AttributeError: pass
            else:
                if isinstance(rewards, str):
                    yield '  Rewards: %s' % rewards
                else:
                    yield '  Rewards:'
                    for r in rewards:
                        yield '    %s' % dumpYamlWithIndent(r, indent = 4)

        return '\r\n'.join(description()) + '\r\n'

    class Activation(Senario.Activation):
        def validate(self):
            Senario.Activation.validate(self)
            self.slayed = []

        # World Events.
        def onSlayMobile(self, victim, killer):
            if victim in self.opponents:
                # What if the killer isn't part of this quest?? (still need to clean up)
                if killer is self.player:
                    self.slayed.append(victim)

                    # Todo: per-opponent slay messages!
                    msg = self.senario.get('slay-message')
                    if msg is not None:
                        peer = self.player.peer
                        if peer is not None:
                            print(msg, file=peer)

                self.opponents.remove(victim)
                self.checkProgress()

        def onDealDamage(self, attacker, victim, dam, wtype):
            # Basically: only allow those involved in the quest to do any damage.
            if victim in self.opponents:
                if attacker is not self.player:
                    # Just regain the damage!
                    victim.hit += dam

        def onPurgeMobile(self, victim, wizard):
            if victim in self.opponents:
                self.opponents.remove(victim)
                self.checkProgress()

        def onDeathTrap(self, victim, cause):
            if victim in self.opponents:
                self.opponents.remove(victim)
                self.checkProgress()

        # House-keeping.
        def deactivate(self, *args, **kwd):
            # Clean up remaining opponents.  Shut down arena.  Burn the remains.
            for opp in self.opponents:
                try: opp.valid
                except ValueError:
                    # Already dead and gone.
                    continue

                # todo: allow corpse-ated rewards.
                opp.act('$n suddenly lights with an inner heat, and then crumbles to dust.')
                eq = opp.equipment
                for o in eq:
                    o = eq[o]
                    if o is not None:
                        o.extract()

                for o in opp.inventory:
                    o.extract()

                # cleanup props
                opp.extract()

            super(self.__class__, self).deactivate(*args, **kwd)

        def checkProgress(self):
            if not self.opponents:
                # Slay senario is done: calculate rewards.
                kitty = self.getKitty()

                qc = QuestControl.get()
                if qc is None:
                    # Fail: dump kitty to disk.
                    pass
                else:
                    # Post kitty to control facility for later redeeming.
                    qc.postReward(self.player, kitty)

                # todo: log victory/defeat
                # todo: announce victory/defeat on the quest channel

                self.deactivate(reason = self.senario.get('end-message', self.senario.DEFAULT_END_MESSAGE))

        def getKitty(self):
            kitty = Kitty()
            for mob in self.slayed:
                opp = self.marks.get(mob)
                if opp is not None:
                    self.addReward(kitty, opp.get('reward'))

            self.addReward(kitty, self.senario.get('reward'))
            return kitty

        def addReward(self, kitty, reward):
            if isinstance(reward, (list, tuple)):
                for reward in reward:
                    if isinstance(reward, dict):
                        for (name, value) in reward.items():
                            if name == 'gold':
                                kitty.addGold(self.pointValue(value))
                            else:
                                point_type = {'qpoints': 'quest',
                                              'quest-points': 'quest',
                                              'silver-points': 'silver',
                                              'gold-points': 'gold',
                                              'platinum-points': 'platinum',
                                              'merit-points': 'merit'}.get(name)
                                if point_type:
                                    kitty.addPoints(point_type, self.pointValue(value))

        def pointValue(self, value):
            if type(value) is int:
                return value

            return evaluate(value, player = self.player) # XXX should this be imported?

        def percentComplete(self):
            return (float(len(self.slayed)) / len(self.opponents)) * 100

    def activate(self, player):
        field = self.get('field', 'world')

        opponents = []
        marks = {}
        for opp in self.get('opponents', []):
            proto = MobilePrototype(opp.get('mark-vnum'))
            for x in range(opp.get('count', 1)):
                mob = self.makeOpponent(player, opp, proto, field)
                marks[mob] = opp
                opponents.append(mob)

        super(self.__class__, self).activate(player, opponents = opponents,
                                             marks = marks)

    VALID_NPCFLAGS = set(['sentinel'])
    VALID_AFFECTFLAGS = set()
    DATA_OVERRIDES = ['level', ('hit-points', 'max_hit'), 'alignment',
                      'experience', 'hitroll', 'damroll', ('armorclass', 'ac'),
                      ('gold', 'gold_on_hand')]

    MAX_LEVEL = 105
    MAX_EXPERIENCE = 10 * 1000 * 1000

    def setFlags(self, bitv, valid, value):
        if isinstance(value, str):
            for f in value.split():
                f = f.lower()

                value = True
                if f[0] == '-':
                    f = f[1:]
                    value = False

                if f in valid:
                    setattr(bitv, f, value)

    def makeOpponent(self, player, opp, proto, field):
        mob = None

        # Select Field.
        if field == 'world':
            for mob in proto:
                break # selecting first.
            else:
                raise RuntimeError('Not in world: %s' % proto)

        elif field == 'roaming':
            mob = proto.instantiate(RealRoom(opp.get('load-room')))
        elif isinstance(field, list):
            try: arena = field[0]['arena']
            except (KeyError, IndexError):
                raise ValueError('Unknown field: %r' % field)

            if arena:
                # arena: load room for mobs, creates self-closing portal.
                # (or just transports).. random arena in arena-group..
                raise NotImplementedError('Arena: %r' % arena)

        assert mob is not None
        npcflags = mob.npcflags

        npcflags.NoCorpse = True # generally desireable
        npcflags.Quest = True

        self.setFlags(npcflags, self.VALID_NPCFLAGS, opp.get('flags'))
        self.setFlags(mob.affectflags, self.VALID_AFFECTFLAGS, opp.get('affect-flags'))

        # Data overrides.
        for name in self.DATA_OVERRIDES:
            if isinstance(name, (tuple, list)) and len(name) == 2:
                (name, attribute) = name
            else:
                attribute = name

            value = opp.get(name)
            if type(value) is int:
                setattr(mob, attribute, value)
            elif value is not None:
                setattr(mob, attribute, evaluate(value, player = player, opponent = mob)) # XXX should this be imported?

        # Note of arrival and equip.
        for msg in opp.get('arrival-messages', []):
            mob.act(msg, object = mob)

        for group in opp.get('equipment', []):
            for (wear_pos, vnum) in group.items():
                item = ItemPrototype(vnum).instantiate(mob)
                mob.equip(item, wear_pos)

        for vnum in opp.get('inventory', []):
            ItemPrototype(vnum).instantiate(mob)

        # todo: make follower chain (using yaml reference syntax, or naming)

        # Normalize:
        mob.level = min(self.MAX_LEVEL, max(0, mob.level))
        mob.hit = mob.max_hit
        mob.experience = min(self.MAX_EXPERIENCE, min(0, mob.experience))
        mob.hitroll = max(0, mob.hitroll)
        mob.damroll = max(0, mob.damroll)

        return mob



# Execute routine system commands and redirect to peer.
from stuphos.runtime.facilities import Facility
from .config import loadConfig
from ..etc import parseOptionsOverSystem
from stuphos import getSection, enqueueHeartbeatTask

import os
from _thread import start_new_thread as nth

class SystemShell(Facility):
    NAME = 'System::Shell'

    class Manager(Facility.Manager):
        VERB_NAME = 'system-sh*ell'
        MANAGE_LEVEL = Facility.Manager.IMPLEMENTOR

        USAGE = 'Usage: system-shell { load | unload | show [--long] | <command> }'

        def doCommand(self, peer, cmd, args):
            if peer.avatar and peer.avatar.level >= self.MINIMUM_LEVEL:
                largs = args and args.strip().lower() or ''
                args = args and args.split() or ()

                try:
                    # Re-mapping facility-management sub-commands:
                    if not largs:
                        print(self.facility.get() or 'Not installed.', file=peer)
                    elif largs == 'load':
                        print(self.facility.get(create = True), file=peer)
                    elif largs == 'unload':
                        print(self.facility.destroy() and \
                              'Destroyed.' or 'Unknown facility.', file=peer)

                    else:
                        this = self.facility.get()
                        if this is None:
                            print('Facility is not loaded.', file=peer)
                        elif not self.doSubCommand(this, peer, args):
                            print(self.USAGE, file=peer)

                except RuntimeError as e:
                    print(e, file=peer)
                except SystemExit:
                    pass

                return True

        def doSubCommand(self, mgr, peer, args):
            name = args[0]
            args = args[1:]

            methname = 'do_%s' % (name,)
            subcmd = getattr(self, methname, None)

            if callable(subcmd):
                subcmd(mgr, peer, args)
                return True

            try: sh = mgr.getShellCommand(name)
            except NameError:
                print('Unknown subcommand or shell command: %r' % name, file=peer)
            else:
                if sh:
                    sh.invoke(mgr, peer, args)
                    return True

        def do_show(self, mgr, peer, args):
            peer.page_string('\r\n'.join('%s: %s' % (name, sh.getSystemCommand()) \
                                         for (name, sh) in mgr.commands.items()) + '\r\n')

        def do_reload(self, mgr, peer, args):
            mgr.reloadAllConfigs()
            print(mgr, file=peer)

    class ShellCommand:
        @classmethod
        def parseConfig(self, section):
            for opt in section.options():
                yield (opt, section.get(opt))

        def __init__(self, name, config):
            self.name = name
            self.config = config

        def getSystemCommand(self):
            try: return self.config['command']
            except KeyError:
                raise ValueError

        def invoke(self, mgr, peer, args):
            ##    if not peer.avatar or peer.avatar.level < self.config.get('level', mgr.Manager.MINIMUM_LEVEL):
            ##        return

            try: systemCommand = self.getSystemCommand()
            except ValueError:
                mgr.syslog('Command not configured: %r' % self.name)
            else:
                mgr.syslog('Executing: %r' % self.name)
                mgr.syslog('...%s' % systemCommand)

                pipe = os.popen(systemCommand)
                def processOutput():
                    enqueueHeartbeatTask(peer.page_string, pipe.read())

                nth(processOutput, ())

        def invokeHeadless(self, mgr, args):
            systemCommand = self.getSystemCommand()
            mgr.syslog('Executing: %r' % self.name)
            mgr.syslog('...%s' % systemCommand)

            return os.popen(systemCommand)

    @classmethod
    def getAllConfigs(self):
        section = getSection('SystemShell')
        if section:
            for opt in section.options():
                if opt == 'path' or opt.startswith('path.'):
                    cfg = loadConfig(section.get(opt))
                    for name in cfg.sections():
                        section = cfg.getSection(name)
                        shellcmdcfg = self.ShellCommand.parseConfig(section)
                        shellcmdcfg = dict(shellcmdcfg)
                        yield (name, self.ShellCommand(name, shellcmdcfg))

    @classmethod
    def create(self):
        return self(dict(self.getAllConfigs()))
    def __init__(self, commands):
        self.commands = commands

    def getShellCommand(self, name):
        try: return self.commands[name]
        except KeyError:
            raise NameError(name)

    def reloadAllConfigs(self):
        # Hard
        self.commands = dict(self.getAllConfigs())

    def syslog(self, message):
        from stuphos.system.api import syslog
        syslog('[%s]: %s' % (self.__class__.__name__, message))

    def __str__(self):
        return '%s (%d commands)' % (self.__class__.__name__,
                                     len(self.commands))
    __repr__ = __str__

# Install Management.
SystemShell.manage()

##    @ACMD('shell-pstree*')
##    def doShellPSTree(peer, cmd, argstr):
##        if peer.avatar and peer.avatar.level >= 115:
##            peer.page_string(os.popen('ps ux').read())
##            return True

# API
try: from stuphos.runtime.registry import registerObject, ObjectAlreadyRegistered
except ImportError: pass
else:
    class API:
        def getSystemShell(self):
            return SystemShell.get()
        def reloadCommands(self):
            return self.getSystemShell().reloadAllConfigs()

        def doCommand(self, name, peer, *args):
            mgr = self.getSystemShell()
            sh = mgr.getShellCommand(name)
            if sh:
                sh.invoke(mgr, peer, args)

        def doCommandHeadless(self, name, *args):
            mgr = self.getSystemShell()
            sh = mgr.getShellCommand(name)
            if sh:
                return sh.invokeHeadless(mgr, args)

    # Singleton.
    API = API()

    SYSTEMSHELL_API_OBJECT_NAME = SystemShell.NAME + '::API'
    try: registerObject(SYSTEMSHELL_API_OBJECT_NAME, API)
    except ObjectAlreadyRegistered:
        pass



# Copyright 2011 Clint Banis.  All rights reserved.
#
from .comm import TelnetConversation, SuMessageBuilder

# Event Handling (Automatic Processing).
# import mud

#@on.telnetCommand
#@mud.tools.breakOn
def process_command(peer, cmd):
    try: conversation = peer.telnet_conversation
    except AttributeError:
        conversation = peer.telnet_conversation = TelnetConversation()

    # This is where to handle other commands.
    ##    if type(cmd) is tuple:
    ##        # Filter for non-SE messages.
    ##        if len(cmd) == 2:
    ##            pass

    msg = conversation.receive_command(cmd)
    if msg is not None:
        return process_message(peer, msg)

def process_message(peer, msg):
    handler = getattr(peer, 'process_telnet_message', process_message_default)
    if callable(handler):
        return handler(peer, msg)

def process_message_default(peer, msg):
    try:
        # Put message in queue if avatar is accepting them.
        try: peer.avatar.telnet_messages.append(load_pickle(msg))
        except AttributeError:
            # Otherwise, just page the message to the player.        
            peer.page_string(str(load_pickle(msg)))

    except:
        from traceback import print_exc
        print_exc(file = peer)


# Manual Processing.
def process_peer(peer):
    builder = SuMessageBuilder()
    while True:
        data = peer.dequeue_input_block()
        if data is None:
            break

        msg = builder.receive_sudata(data)
        if msg is not None:
            return msg

def process_player(player):
    # (for deferred processing when everything's put
    # on a queue for later debugging)
    assert player.peer
    messages = getattr(player, 'telnet_messages', None)
    if messages is None:
        messages = player.telnet_messages = []

    msg = process_peer(player.peer)
    if msg is not None:
        messages.append(msg)



# Python Code Shell Evaluator.
import stuphos
from stuphmud import server as mud
from stuphos.etc.tools import getSystemException, Attributes
from stuphmud.server.player import HandleCommandError, policy, getPlayerScope
from stuphmud.server.player.interfaces.code import ShellFactoryBase
from . import ProgrammeManager, EvaluationShell, TemporaryCodeFile, isFromSecureHost

@Attributes.Dynamic
def Environment(player, name, **search):
    return player.find(name, **search)

class PythonCodeManager(ProgrammeManager):
    # ProgrammeManager:
    def getManagerName(self):
        return 'Python'
    def getManagerId(self):
        return '%s.%s.Instance' % (self.__class__.__module__, self.__class__.__name__)

    # Implementation.
    isFromSecureHost = staticmethod(isFromSecureHost)
    def isPythonAllowed(self, peer):
        # Connection must originate from this machine.
        if not self.isFromSecureHost(peer):
            return False

        avatar = peer.avatar
        if avatar:
            if avatar in policy:
                # Powerful.
                return True

            # This line allows all implementors to use python:
            return avatar.supreme

        return False

    PROLOG = 'from __future__ import with_statement\n'
    EPILOG = ''

    def getDefaultShellName(self):
        # (code source)
        return '?'

    def compileSourceCodeBlock(self, peer, sourceCode, shellName = None):
        # Remove-preceding whitespace.
        sourceCode = self.formatSourceCode(sourceCode.lstrip())
        sourceCode = ''.join((self.PROLOG, sourceCode, self.EPILOG))

        if shellName is None:
            shellName = self.getDefaultShellName()

        # Compile code from source.
        with TemporaryCodeFile(peer, sourceCode, shellName):
            return compile(sourceCode, shellName, 'exec')

    def compileSingleStatement(self, peer, sourceCode, shellName = None):
        if shellName is None:
            shellName = self.getDefaultShellName()

        sourceCode = self.formatSourceCode(sourceCode)
        with TemporaryCodeFile(peer, sourceCode):
            return compile(sourceCode, shellName, 'single')

    def executeCode(self, shell, peer, code, **environ):
        # Hack: inspect evaluator stack on CodeManager Shell.
        managerName = self.getManagerName()

        try: execute = shell.getEvaluatorByManagerName(managerName).executeCode
        except AttributeError: pass
        else: return execute(shell, peer, code, **environ)

class PythonCodeShell(EvaluationShell):
    # EvaluationShell:
    def executePython(self, shell, peer, argstr):
        if argstr and self.manager.isPythonAllowed(peer):
            return self.executeSourceCode(shell, peer, argstr)

    __call__ = executePython

    def executeCode(self, shell, peer, code, **environ):
        local = self.setupLocalNamespace(shell, peer, (None, None), **environ)

        ##    if getattr(self, 'debug', False):
        ##        debug_code(code, local)

        try: exec(code, self.global_namespace, local)
        except:
            # For now, show all how we got here.
            HandleCommandError(peer, getSystemException(), frame_skip = 0)

    # Implementation.
    thisName   = 'this' # avatar's peer
    myName     = 'me'   # peer's avatar
    myLocation = 'here' # avatar's room location

    shellName  = '<player>'

    def __init__(self, manager, scope):
        EvaluationShell.__init__(self, manager)
        self.scope = scope

        # I guess these are separately-configurable.
        self.local_namespace = scope.__dict__
        self.global_namespace = getMainNamespace(scope)

    def getShellName(self, peer):
        'Code Source (not source code).'

        self.compiler_name = shellName = getattr(peer, 'host', False) or self.shellName
        return shellName

    def compileSourceCodeBlock(self, peer, sourceCode):
        return self.manager.compileSourceCodeBlock(peer, sourceCode, self.getShellName(peer))
    def compileSingleStatement(self, peer, sourceCode):
        return self.manager.compileSingleStatement(peer, sourceCode, self.getShellName(peer))

    def setupLocalNamespace(self, shell, peer, xxx_todo_changeme, **environ):
        (cmd, argstr) = xxx_todo_changeme
        local = self.local_namespace
        avatar = peer.avatar

        local.update({
            self.thisName   : peer,
            self.myName     : avatar or peer.avatar,
            self.myLocation : avatar and avatar.room,

            # '__command__'   : cmd,
            # '__argstr__'    : argstr,

            'commands'      : shell.commands,
            'acmd'          : getattr(shell.commands, 'insert', None), # or just self.insert
            'history'       : shell.historyObject(peer),
            'last'          : lambda *args, **kwd:shell.historyLast(peer, *args, **kwd),

            'sh'            : shell,
            'run'           : shell.enqueueHeartbeatTask,
            'spawn'         : shell.spawnTask,
            'inline'        : stuphos.enqueueHeartbeatTask,

            'e'             : Environment(avatar),
            'g'             : Environment(avatar, mobile_in_world = True, item_in_world = True),

            # 'com'           : self.getCode,
        })

        local.update(environ)
        return local

# Infrastructure.
def getMainNamespace(scope):
    main = getModule()
    main.shell = scope
    main.main = main

    from stuphmud import server as mud
    import pdb
    main.mud = mud
    main.pdb = pdb

    try: from stuphos.system.api import game, world
    except ImportError: pass
    else:
        main.game = game
        main.world = world

    import builtins as builtin
    main.builtin = builtin

    import sys, os
    main.sys = sys
    main.os = os

    return main.__dict__

def getModule(name = '__main__'):
    # This should go in tools.

    try: main = __import__(name)
    except ImportError:
        main = new.module(name)
        from sys import modules
        modules[name] = main

    return main

# Singleton.
PythonCodeManager.Instance = PythonCodeManager()

def getPythonCodeEvaluator(scope):
    return PythonCodeShell(PythonCodeManager.Instance, scope)

class PythonCodeEvaluatorFactory(ShellFactoryBase):
    def __new__(self, peer):
        print(EvaluationShell.PROGRAMMING_HEADER % 'Python', file=peer)
        return getPythonCodeEvaluator(getPlayerScope(peer))



from sqlobject import *

class StuphRoomRecord(SQLObject):
	vnum            = IntCol()

	name            = StringCol()
	descr           = StringCol()
	flags           = IntCol()
	sector          = IntCol()

	exits           = BLOBCol()
	extradesc       = BLOBCol()

	stuphZoneRecord = ForeignKey('StuphZoneRecord')




# class LibraryDispatch(SubcommandMode):
# 	def __init__(self, path, previous = None):
# 		BaseMode.__init__(self, previous = previous)
# 		self.path = path

#     def defaultCommand(self, engine, peer, name, *args, **kwd):
#     	# get programmer setting from authenticated peer association.
#     	# run in a separate thread, and wait for the task completion to post result for return.
#     	pass

# # Craft an application that pushes LibraryDispatch modes based on configuration.
# # One way to do this is to override Application.LoggedInMode.

# from receptacle.architecture import ServiceBase
# class Girl(ServiceBase):
#     NAME = 'Agent'

#     def doRunModule(self, engine, peer, source, *args, **kwd):
#         from stuphos.kernel import newModuleTask
#         from Queue import Queue

#         q = Queue()

#         @newModuleTask(source, arguments = args, **kwd)
#         def response(result):
#         	q.put(result)

#         return q.get()

#     def doEvaluate(self, engine, peer, source, *args, **locals):
#         from stuphos.kernel import Girl, Script, executeGirl
#         from Queue import Queue

#         q = Queue()

#         procedure = Girl(Girl.Expression, source)
#         locals['arguments'] = args

#         task = Script.Load()
#         task.addFrameCall(procedure, locals = locals)

#         @executeGirl(task)
#         def response(task):
#             # tag:task-completion
#             q.put(task.stack[0])

#         return q.get()

#     def doRunAssembly(self, engine, peer, source, *args, **locals):
#         from stuphos.kernel import Girl, Script, executeGirl

#         procedure = Girl(Girl.Asm, source)
#         locals['arguments'] = args

#         task = Script.Load()
#         task.addFrameCall(procedure, locals = locals)

#         executeGirl(task)

#     def doRunMachineCode(self, engine, peer, code, *parameters, **locals):
#         from stuphos.kernel import Girl, Assembly, Script, executeGirl

#         # Compile and run.
#         procedure = Girl(Assembly, code)
#         locals['arguments'] = args

#         task = Script.Load()
#         task.addFrameCall(procedure, locals = locals)

#         executeGirl(task)

#     def doOpenServerChannel(self, engine, peer, name, *args, **kwd):
#         from stuphos.kernel import serverChannel

#         core = runtime[runtime.Agent.System]
#         assert core is not None

#         return serverChannel(core, name)


# class Federation(ServiceBase):
#     NAME = 'Federation'

#     def doRouteCall(self, engine, peer, routeAddress, method, *args, **kwd):
#     	pass



'''
[Mobiles:person.features.specproc]
3039: healer.Healer

[Rooms]
0: Dump

'''

from stuphos.runtime import LookupClassObject
from stuphos.runtime.facilities import Facility
# from mud.tools.logs import logException
from stuphos.etc import splitOne
from stuphos import logException

from stuphos.system.api import syslog, world
BuiltinSpecialProcedure = world.special

#from traceback import print_exc as traceback
from configparser import ConfigParser

# Special Procedures Management.
class SpecprocBitvector(int):
    pass

Room   = SpecprocBitvector(1 << 0)
Mobile = SpecprocBitvector(1 << 1)
Item   = SpecprocBitvector(1 << 2)

Any    = SpecprocBitvector(Room | Mobile | Item)

class Table:
    def __init__(self, parent = None, name = None):
        self.parent = parent
        self.name = name

        # What about by name?
        self.rooms = {}
        self.mobiles = {}
        self.items = {}

        self.subtables = {}
        if parent:
            parent.registerSubtable(self, name)

        # Do this dynamically (shy from direct access without mud api)
        from world import room, mobile, item
        self.RealRoom = room
        self.RealMobile = mobile
        self.RealItem = item

    def assign(self, object, room = None,
               mobile = None, item = None):

        self.assignRoom(room, object)
        self.assignMobile(mobile, object)
        self.assignItem(item, object)

    def register(self, object, room = None,
                 mobile = None, item = None):

        self.registerRoom(room, object)
        self.registerMobile(mobile, object)
        self.registerItem(item, object)

    def registerSubtable(self, table, name):
        if name:
            self.subtables[name] = table

    # Registration: Put procedure into the table, associated with
    # name or Vnum information.
    def registerRoom(self, room, object):
        if type(room) is int:
            self.rooms[room] = object
            return self.RealRoom(room)

        if type(room) is world.room:
            # recognize re-register
            self.rooms[room.vnum] = object
            return room

    def registerMobile(self, mobile, object):
        if type(mobile) is int:
            self.mobiles[mobile] = object
            return self.RealMobile(mobile)

        if type(mobile) is world.mobile:
            # recognize re-register
            self.mobiles[mobile.vnum] = object
            return mobile

    def registerItem(self, item, object):
        if type(item) is int:
            self.items[item] = object
            return self.RealItem(item)

        if type(item) is world.item:
            # recognize re-register
            self.items[item.vnum] = object
            return item

    # Assignment: The interface-with-core operation of
    # procedure installation.
    def assignRoom(self, room, object, **values):
        try: room = self.registerRoom(room, object)
        except ValueError:
            syslog('SYSERR: Assigning %r to unknown room: %s' % (object, room))
        else:
            if room is not None:
                # todo: recognize overwrite
                DoBindSpecial(room, object, **values)

    def assignMobile(self, mobile, object, **values):
        try: mobile = self.registerMobile(mobile, object)
        except ValueError:
            syslog('SYSERR: Assigning %r to unknown mobile: %s' % (object, mobile))
        else:
            if mobile is not None:
                # todo: recognize overwrite
                DoBindSpecial(mobile, object, **values)

    def assignItem(self, item, object, **values):
        try: item = self.registerItem(item, object)
        except ValueError:
            syslog('SYSERR: Assigning %r to unknown item: %s' % (object, item))
        else:
            if item is not None:
                # todo: recognize overwrite
                DoBindSpecial(item, object, **values)

def DoBindSpecial(entity, object, **values):
    special = getBoundSpecial(entity, object, **values)
    if special is not None:
        entity.special = special

    return special

def getBoundSpecial(entity, object, **values):
    try: bind = object.BindSpecial
    except AttributeError: return object

    if not callable(bind):
        # I guess we'll just ignore this attribute.
        return object

    # Do dynamic binding of specproc to entity.
    result = bind(entity, **values)
    if result is not False:
        if callable(result):
            return result

        return object

# Builtin Table.
class SpecialManager(Facility, Table):
    NAME = 'SpecialProcedure::Manager'

    @classmethod
    def create(self):
        table = self(name = '')
        for name in ('assign', 'register', 'assignItem',
                     'assignMobile', 'assignRoom'):

            # Basically provide a recapitalized form of the attribute.
            setattr(self, name[0].upper() + name[1:],
                    staticmethod(getattr(table, name)))

        return table

# Install.
# SpecialManage.manage()
getBuiltinTable = SpecialManager.get
getBuiltinTable(create = True)

# Configuration Layer.
def parseSpecialsConfigSection(section):
    parts = section.split(':', 1)
    if len(parts) == 2:
        return parts
    if len(parts) == 1:
        return (parts[0], None)
    return (None, None)

def LookupSpecialObject(name, namespace = None, spec_type = None):
    if namespace:
        name = '%s.%s' % (namespace, name)
    elif spec_type and '.' not in name:
        try: return BuiltinSpecialProcedure(name, spec_type)
        except NameError:
            return None

    try: return LookupClassObject(name, raise_import_errors = True)
    except (ImportError, AttributeError):
        logException(traceback = True)

    # This raised from LookupObject, which doesn't differentiate
    # types.  So, provide our own type by recognizing this error.
    ##    except AssertionError:
    ##        raise UnknownSpecialObject(name, namespace, spec_type)


def parseSpecialsConfigType(config, special_type):
    results = []
    type = special_type.lower()
    for section in config.sections():
        (type, namespace) = parseSpecialsConfigSection(section)
        if type.lower() == special_type:
            for opt in config.options(section):
                try: vnum = int(opt)
                except ValueError:
                    continue

                values = config.get(section, opt)
                values = values.split('\n')
                name = values[0]

                if len(values) == 1:
                    values = dict()
                else:
                    # Configure binding parameters.
                    from yaml import load
                    values = load('\n'.join(values[1:]))

                results.append((vnum, namespace, name, values))

    return results

def parseSpecialsConfig(config):
    return dict(rooms = parseSpecialsConfigType(config, 'rooms'),
                mobiles = parseSpecialsConfigType(config, 'mobiles'),
                items = parseSpecialsConfigType(config, 'items'))

def parseSpecialsConfigFromFile(config_file):
    config = ConfigParser()
    config.read([config_file])
    return parseSpecialsConfig(config)

def parseSpecialsConfigFromString(string):
    from io import StringIO
    config = ConfigParser()
    config.readfp(StringIO(string)) # , '<init-buf>')
    return parseSpecialsConfig(config)

# Real World Loading.
def loadSpecialsConfig(config):
    # Todo: maybe log more about which specproc this is happening for.
    for (kind, assign) in [('rooms', SpecialManager.AssignRoom),
                           ('mobiles', SpecialManager.AssignMobile),
                           ('items', SpecialManager.AssignItem)]:

        for (entity, ns, name, values) in config[kind]:
            try: object = LookupSpecialObject(name, ns, kind)
            except NameError as e: syslog('SYSERR: %s (%s)' % (e, e.__class__.__name__))
            except Exception: logException(traceback = True)
            else:
                if object is None:
                    syslog('SYSERR: Unknown %s procedure [%s::%s]' % (kind, ns, name))
                else:
                    try: assign(entity, object, **values)
                    except: logException(traceback = True)

def loadSpecialsConfigFromFile(config_file):
    loadSpecialsConfig(parseSpecialsConfigFromFile(config_file))
def loadSpecialsConfigFromString(string):
    loadSpecialsConfig(parseSpecialsConfigFromString(string))

# todo: loadSpecialsConfigFromDatabase, for in-game-managed assignment file.
# see: mud.player.commands.gc.wizard [ControlPoints::assign]

# Object Model.
def isCommandReserved(cmd):
    return cmd is None or cmd.name == 'RESERVED'
def parseSpecialCommand(argstr):
    if argstr is None:
        return ('', '')

    return splitOne(argstr, ' ')


# import standard



# BOARD - Basic Object And Response Directory
'''
Content Management System with persistant storage,
a directory structure, and an interactive special
procedure interface.

Produces RFC288 (MIME) email messages.

Framework for generating all sorts of structured
data from the world/connected programs.


[Item]
1282: ph.lang.board.PropertySheet
1283: ph.lang.board.CommandMessage
    action = stuph/x-board-reimbursement

; Custom Data View.
1284: ph.lang.board.DataTemplate

[Room]
; Immortal Boardmaster
1280: ph.lang.board.repository
    item-class = mud.management.board.Post
    smtp-host: stuphmud.net
    ; smtp-port: 20
    path = plrmail/admin-repos
    prototype = 1281

    ; GC:
    ; at 1280 check out fraun/cainus/aug8th2014
    ; pload self 1282
    ; @me('set properties:message-signature %s --method=sha' % \
    ;     calc.sha(json.dumps(map(attributeOf.store, me.contents))))
    ; @set properties:message-signature --method=sha \
    ;     {calc.sha(json.dumps(map(attributeOf.store, me.contents)))}
    ; put properties package
    ; put all package
    ; pload self 1283
    ; put reimbursement package
    ; close package
    ; mail package cainus@stuphmud.net

    '''

try: from stuphmud.server.zones.specials.standard import Special
except Exception as e:
    print(e)
    class Special:
        pass

# Todo: remove dependency on this module by making closeCommand
# initialization dynamic.
import world

try: from common.runtime.virtual.objects import attributeFor
except ImportError:
    # Remove too much reliance on wrlc.
    class attributeFor:
        def __init__(self, name):
            self.name = name
        def __call__(self, object):
            return getattr(object, self.name)

from contextlib import contextmanager

specialOf = attributeFor('__special__')
closeCommand = world.command('close')

class BOARDRepository(Special):
    COMMANDS = dict(check = 'doCheck')

    def doCheck(self, player, this, argstr):
        args = argstr.split()
        if len(args) > 0 and args[0] == 'out':
           # and this is player.find(args[1]):
            item = self.BOARDItemClass.CreateFor(self, args[1], player)
            player.act('You check out $o from $O', FALSE, item, this)

            return True

    class BOARDItem(Special):
        @classmethod
        def CreateFor(self, bRepos, name, player):
            item = world.item(self.getMeta(args[0])['prototype'])

            assert specialOf(item) is None or \
                   isinstance(specialOf(item), self)

            item = item.instantiate(player)
            item.__special__ = self(bRepos, name)

            return item

        COMMANDS = dict(check = 'doClosePackage') # XXX should this be named 'check'?

        def doClosePackage(self, player, this, argstr):
            # Commit contents.
            # todo: parse argstr for --force --message parameters
            args = argstr.split()

            if len(args) == 1:
                name = args[0]

                if this is player.find(name):
                    if this.isClosed():
                        closeCommand(player, name)
                    else:
                        try: self.commit(player, this)
                        except:
                            player.peer.handleSystemException()
                            closeCommand(player, name)
                        else:
                            # todo: mudlog
                            closeCommand(player, name)
                            if player.peer:
                                print('&N* &wCommitted&N *', file=player.peer)

                    return True

        def __init__(self, bRepos, name):
            self.board = bRepos(name)

        @contextmanager
        def wc(self, player):
            'Manage a temporary working copy file in player folder.'

            path = getPlayerFilename(player) # ...
            wc = self.board.checkout(path)

            try: yield (wc, path)
            finally: path.delete() # or something

        def commit(self, player, this):
            contents = BOARDPackStuphOSItem(this)

            with self.wc(player) as (wc, path):
                path.write(contents)
                response = wc.commit('Update {timestamp}')

            # It occurs to me that this is useless for most applications.
            ##    for o in this.contents:
            ##        # Report errors, but proceed completely with commit.
            ##        try: o.extract()
            ##        except: player.peer.handleSystemException()

            return response

    BOARDItemClass = BOARDItem

    def __init__(self, path, **values):
        self.repos = SVNModule.REPOS.FromLocalPath(path)
        self.values = values

        try: itemClass = values['item-class']
        except KeyError: pass
        else: self.BOARDItemClass = LookupObject(itemClass)


# For easier configuration (see doc):
repository = BOARDRepository
repository.item = repository.BOARDItem

class SVNModule:
    # my.programmes.interface.svn.wmcOf
    class REPOS:
        @classmethod
        def FromLocalPath(self, path, api = None):
            if api is None:
                api = defaultSvnApi

            return self(api, io.path(path).urlString)

        def __init__(self, api, uri):
            self.api = api
            self.uri = uri

        def __call__(self, name):
            return call(classOf(self),
                        '%s/%s' % (self.uri,
                                   name))

        def checkout(self, path):
            return self.WC(self, path).checkout()

        class WC:
            def __init__(self, repos, path):
                self.repos = repos
                self.path = path

            def checkout(self):
                return self


# BOARD Packaging Protocol.
BOARDPackStuphOSItemName = 'StuphOS_BOARD_Pack'

def BOARDPackStuphOSItem(item):
    msg = buildMessage('stuph/x-board-package')

    for o in item.contents:
        try: pack = getattr(specialOf(o), BOARDPackStuphOSItemName)
        except AttributeError:
            # WTF special case.
            try: action = o.XSTUPH_ACTION
            except AttributeError: pass
            else: return buildMessage \
                  (**{'X-Stuph-Action':
                      action})

            # todo: charset.
            msg.attach(buildMessage(content = str(buffer(o.store))))
        else:
            r = pack(o)
            if r is None:
                continue

            if isinstance(r, str):
                r = buildMessage(content = r)

            msg.attach(r)

    return msg

setattr(BOARDRepository.BOARDItem,
        BOARDPackStuphOSItemName,
        staticmethod(BOARDPackStuphOSItem))


# MIME
from email.message import Message

def buildMessage(type = None, content = None, **values):
    msg = Message()
    msg.set_type(type or msg.get_content_type())
    # msg.headers.update(values)

    if content is list:
        for c in content:
            if isinstance(c, list):
                c = buildMessage(content = c)
            elif isinstance(c, str):
                c = buildMessage(content = c)

            msg.attach(c)

    else:
        # set_charset?
        msg.set_payload(content)

    for (n, v) in iteritems(values):
        if isinstance(v, (list, tuple)):
            (v, p) = v

        elif isinstance(v, dict):
            v = ''
            p = v

        else:
            p = dict()

        msg.add_header(n, v)

        for (pn, pv) in iteritems(p):
            msg.set_param(pn, pv, header = n)

    return msg


# Interface Implementations.
class Post(repository.item):
    # Mail
    '''
    mail spec-proc:
        adds MIME message objects to multi-part message
        also holds meta data about editing process:
            date/time started editing
            intended recipients
            date/time sent
            other certificates

        message content:
            resource packages
            command-messages, utilizing resource packages
            guid per player
            gleaned conversational/board topics
            awareness of attached item

        has backing file/db store,
            keyed on guid
            "shared" in directory
            this guid attached to item entity handle
            and/or, attached to the procedure instance
            svn repository??

        spec-proc implementation means can be attached
        to any object, which can be aware of item and
        package it in the message.

        "send" or "share"

    directory structure:
        separately-configured repository per item/room

        manages MIME message files
        one file-per-item
        WC backing-store (temporary.. will need a persistant map)
        libsvn bindings??

        'export' operation to file-backed (notebook) store
        then, import operation of notebook into new message
            subdirectory
            '''

    COMMANDS = dict(repository.item.COMMANDS,
                    mail = 'doMailPackage')

    def doMailPackage(self, player, this, argstr):
        args = argstr.split()
        if len(args) > 1: # at least this target and one recipient.
            if this is player.find(args[0]):
                if player.isPlayer:
                    from smtplib import SMTP
                    recipients = args[1:]

                    msg = str(BOARDPackStuphOSItem(this))

                    try:
                        # Todo: enqueue for another thread to process,
                        # but also mark this object with the enqueued
                        # mailing process, so that it gets notified
                        # on completion.
                        smtp = SMTP(self.bRepos.values['smtp-host'])
                        errors = smtp.sendmail('player-%s@stuphmud.net' % \
                                               player.name, recipients,
                                               msg)
                    except Exception as e:
                        if player.peer:
                            print('%s: %s' % \
                                  (nameOf(classOf(e)), e), file=player.peer)
                    else:
                        if player.peer:
                            player.peer.page(inspect(errors))

                    smtp.quit()
                    return True


# Other Content Items.
class CommandMessage:
    @classmethod
    def BindSpecial(self, item, **values):
        item.XSTUPH_ACTION = values['action']

class DataTemplate:
    '''
    player data:

    {{ data.owner.record }}

    '''

    class StuphOS_BOARD_Pack(object):
        def __new__(self, item):
            from django.template import Template, Context

            return Template(item.detailed_description) \
                   .render(Context(dict(data = object.__new__ \
                                        (self, item))))

        def __init__(self, item):
            self._item = item

        @property # .memorized
        def owner(self):
            o = self._item
            while o is not None:
                if type(o) is world.mobile:
                    return self.Player(o)

                o = o.location

        @property
        def item(self):
            return self.Item(self._item)

        class Accessor:
            def __init__(self, object):
                self._object = object

            @property
            def record(self):
                return str(buffer(self._object.store)) \
                       .encode('base64')

        class Player(Accessor):
            pass
        class Item(Accessor):
            pass


class PropertySheet(Special, dict):
    COMMANDS = dict(set = 'doSetValue',
                    look = 'doLookAt')

    def doSetValue(self, player, this, argstr):
        # todo: world lookup/table results, mailable to self

        # set board-object:x-header-name = value
        # set board-object:x-header-name value rest

        # set board-object x-header-name value rest
        # set board-object x-header-name --parameter=value

        # set board-object:x-header-value
        # ] .

        args = argstr.split()
        if args:
            name = args[0]
            i = name.find(':')

            if i < 0:
                if not args:
                    return False

                header = args[1]
                args = args[2:]

            else:
                header = name[i+1:]
                name = name[:i]

                args = args[1:]

            if args[0] == '=':
                args = args[1:]

                if not args:
                    return False

            (parameters, args) = self.parseArgumentParameters(args)

            if args:
                args = ' '.join(args)

                try: s = self[header]
                except KeyError:
                    s = (args, [])
                else:
                    s = (args, s[1])

                if parameters:
                    s[1].extend(parameters)

                self[header] = s

            elif parameters:
                s = self.setdefault(header, ('', []))

                u = dict(s[1])
                u.update(dict(parameters))

                s = (s[0], list(iteritems(u)))

                s[header] = s

            else:
                return True # todo: start editing string

            if player.peer:
                print(player.peer, 'Set.')

            return True

    def parseArgumentParameters(self, args):
        p = []
        r = []

        for a in args:
            if a[:2] == '--':
                a = a[2:]
                i = a.find('=')

                if i < 0:
                    p.append((a, True))
                else:
                    p.append((a[:i], a[i+1:]))

            else:
                r.append(a)

        return (p, r)

    def doLookAt(self, player, this, argstr):
        args = argstr.split()
        if len(args) == 1 and this is player.find(args[0]):
            if player.peer:
                # Render preview.
                player.peer.page(str(self.StuphOS_BOARD_Pack(this)))
                return True


    def StuphOS_BOARD_Pack(self, item):
        return buildMessage(content = item.detailed_description
                            **self)


def loadServicesBOARDFromStuphLibBuffers(board, buffers):
    import datetime

    # boards.cpp:Board_load_board
    # Load the damn stuphlib.boards.ReadBoardFile vector
    # into .board.specproc objects.
    for b in buffers:
        msg = board.Message(board, b.header, # b.info.heading
                            timestamp = datetime.datetime \
                            .fromtimestamp(b.info.birth),
                            author_level = b.info.level)

        msg.body = b.message
        board.messages.append(msg)

        # yield msg

    return board



# The lang package.  This should be renamed to 'interface', and import certain format-related
# packages from component (like structural).
#
# The 'layer' package can be renamed to the generic 'language', as a layer nomination is better
# suited to the runtime package.  The web package could be considered a part of the interface
# package, but because it represents a core application feature (a server), it isn't a priority.
#
# The purpose is not to generate packages with gigantic numbers of submodules, but rather to
# distributed component functionality evenly across a balanced tree, so that the features of
# the application are rationaly organized and only complex and involved algorithms need take
# up large module sizes.



# System path, module and uptime checkpoint control.
# Copyright 2011 Clint Banis.  All rights reserved.
from stuphos.system.api import mudlog, syslog
from stuphos.etc.tools import maxWidth, columnize

import sys as system_module
from os.path import join as joinpath, isdir, exists, sep as ROOT_SEP, expanduser
from datetime import datetime
from time import time as getCurrentSystemTime

CYGDRIVE = ROOT_SEP + 'cygdrive'
def translateRoot(name):
    if len(name) == 1 and name.isalpha():
        return joinpath(CYGDRIVE, name)
    elif name in '~@':
        return expanduser('~')

    return name

def buildPath(*args):
    assert args
    return joinpath(translateRoot(args[0]), *args[1:])

def isZipArchive(path):
    # lame:
    return path.endswith('.zip') or path.endswith('.egg')

def getPathInfo(path):
    kind = ' '
    if isdir(path):
        kind = '+'
    elif exists(path):
        if isZipArchive(path):
            kind = 'Z'
    else:
        kind = '!'

    return '%s %s' % (kind, path)

def doSystemPathControl(peer, action = None, *args):
    if action == 'show' or not action:
        peer.page_string('\r\n'.join(map(getPathInfo, system_module.path)) + '\r\n')

    elif action == 'register':
        path = buildPath(*args)
        path = path.replace('+', ' ')
        if path in system_module.path:
            print('Already Registered: %s' % path, file=peer)
        else:
            system_module.path.append(path)
            print('Registered: %s' % path, file=peer)

    elif action == 'unregister':
        path = buildPath(*args)
        path = path.replace('+', ' ')
        if path in system_module.path:
            system_module.path.remove(path)
            print('Removed: %s' % path, file=peer)
        else:
            print('Not Registered: %s' % path, file=peer)

def doSystemModuleControl(peer):
    sysModules = system_module.modules
    modules = list(sysModules.keys())
    modules.sort()

    # Find all top-level module packages.
    toplevel = []
    for m in modules:
        for t in toplevel:
            if m.startswith(t):
                break
        else:
            toplevel.append(m)

    # Gather modules under their imported paths.
    builtins = []
    paths = {}
    from os.path import dirname, basename
    for t in toplevel:
        try: f = sysModules[t].__file__
        except AttributeError:
            builtins.append(t)
            continue

        if f is not None:
            if basename(f).startswith('__init__.'):
                # Trully is a package module: use the parent of parent.
                f = dirname(dirname(f))
            else:
                f = dirname(f)

            paths.setdefault(f, []).append(t)

    # Render display of them.
    def recolumnize(a, tab = '  ', min = 3, max = 20):
        return tab + ('\n' + tab).join(columnize(a, min, maxWidth(a, max)).split('\n'))
    def group(name, a):
        return '\n'.join([name + ':', (len(name) + 1) * '=', recolumnize(a)])

    s = []
    for p in system_module.path:
        m = paths.get(p)
        if m:
            s.append(group(p, m))

    if builtins:
        s.append(group('Builtin Modules', builtins))

    peer.page_string('\n\n'.join(s) + '\n')

# Checkpointing (Uptime Tracking).
# Todo: put into web.services, because it's not just command-oriented.
# The command stuff can be registered dynamically with the WizControl API.
DEFAULT_INTERVAL = (60 * 20, 0) # 20 Minutes
SYSTEM_CHECKPOINT_STATUS_NAME = 'System::Checkpointing::Status'
SYSTEM_CHECKPOINT_LAST_NAME = 'System::Checkpointing::LastTime'

from stuphos.runtime.registry import getObject, setObject
def isCheckpointingEnabled():
    return bool(getObject(SYSTEM_CHECKPOINT_STATUS_NAME))
def setCheckpointingEnabled(status = True):
    setObject(SYSTEM_CHECKPOINT_STATUS_NAME, bool(status))

def getLastCheckpoint():
    return getObject(SYSTEM_CHECKPOINT_LAST_NAME)
def setLastCheckpoint(value):
    setObject(SYSTEM_CHECKPOINT_LAST_NAME, value)

def systemCheckpoint():
    # Pings an outside system monitor to let it know we're still running.
    # Todo: do this in another thread, since web communication obviously takes time.accept2dyear
    # The event-queue can handle the callback of success or failure.
    # XXX Where is this?
    from stuphos.network.services import CheckpointUptime
    message = CheckpointUptime()
    if message:
        message = 'CHECKPOINT-ERROR: %s' % message
    else:
        message = 'SYSTEM: Uptime Checkpoint'
        setLastCheckpoint(getCurrentSystemTime())

    mudlog(message)
    syslog(message)

def systemCheckpointAdapter(sec, usec):
    systemCheckpoint()

forceSystemCheckpoint = systemCheckpoint

def EnableCheckpointing(interval = DEFAULT_INTERVAL):
    if not isCheckpointingEnabled():
        from world import heartbeat
        heartbeat.registerRealtimeAdapter(systemCheckpointAdapter, interval or DEFAULT_INTERVAL)
        setCheckpointingEnabled()
        return True

def DisableCheckpointing():
    if isCheckpointingEnabled():
        from world import heartbeat
        heartbeat.unregisterRealtimeAdapter(systemCheckpointAdapter)
        setCheckpointingEnabled(False)
        return True

def doSystemCheckpointControl(peer, action = 'show', interval = None):
    if action == 'show':
        print('Checkpointing %sactivated.' % ('' if isCheckpointingEnabled() else 'de'), file=peer)
        last = getLastCheckpoint()
        if last is not None:
            delta = datetime.fromtimestamp(getCurrentSystemTime()) - datetime.fromtimestamp(last)
            seconds = delta.seconds # total_seconds() -- version incompatibilities
            print('Last Checkpoint: %s ago' % seconds, file=peer) # days, seconds, microseconds
            # what about seconds to next??

    elif action == 'enable':
        print('Checkpointing %senabled.' % ('' if EnableCheckpointing(interval) else 'already '), file=peer)

    elif action == 'disable':
        print('Checkpointing %sdeactivated.' % ('' if DisableCheckpointing() else 'already '), file=peer)

    elif action == 'update':
        if isCheckpointingEnabled():
            forceSystemCheckpoint()
            print('Checkpoint forced.', file=peer)
        else:
            print('Checkpointing deactivated.', file=peer)

    else:
        print('Unknown system-checkpoint action: %r' % action, file=peer)



# Copyright 2011 Clint Banis.  All rights reserved.
# Telnet Message Dispatch.
#
r"\xff\x0fI44.JSON:['CREATE','FILE-SCRIPT.PY','\"Document\"']\xff\x0f"

from stuphmud.server.player import playerAlert, PlayerAlert, getCapitalizedName
from stuphmud.server.player import HandleCommandError as HandleTelnetError

from stuphos.etc.tools import reraiseSystemException
from functools import reduce

class TelnetMessenger:
    def process(self, peer, msg):
        # Peer Head -- Telnet Message
        #
        # Also of note: the result of this method generally propogates back up to
        # the Network::getCommand (EventBridge), and returning < 0 will prevent any
        # further commands from being processed during that cycle.
        try:
            # todo: build in message digest auth here! (while we still have payload)
            msg = self.loadMessage(msg)
            if type(msg) in (list, tuple) and len(msg) > 0:
                # Here: Beware returning this value..? Affects network flow.
                self.dispatch(peer, msg[0], msg[1:])
            else:
                # Or syslog..?  no, back to where it started from.
                playerAlert('Unknown message type/length: %s', type(msg).__name__)

        except PlayerAlert as e:
            e.deliver(peer)
        except:
            # To syslog... This is more of a system error.
            from stuphmud.server.player import HandleCommandError as HandleTelnetError
            HandleTelnetError(peer)

    def loadMessage(self, msg):
        # First dereference the string buffer.
        msg = msg.getvalue()

        if msg.startswith('JSON:'):
            from simplejson import loads
            msg = loads(msg[5:])
        elif msg.startswith('PICKLE:'):
            from pickle import loads
            msg = loads(msg[7:])
        elif msg.startswith('XMLRPC:'):
            from xmlrpc.client import loads
            (params, method) = loads(msg[7:])

            # Flatten parameters into one list.
            msg = [method]
            msg.extend(params)

        elif msg.startswith('PTCL:'): # PNTL? RCTL?
            return ['pentacle:command', msg[5:]]

        elif msg.startswith('HTTP:'):
            playerAlert('Unable to process HTTP request.')

        return msg

    def getDispatchAction(self, operation):
        if self.dispatch_module is not None:
            # First treat the module as a delegate.
            delegate = getattr(self.dispatch_module, 'getDispatchAction', None)
            if callable(delegate):
                return delegate(operation)

            # Otherwise, perform string lookups.
            if type(operation) is not str:
                playerAlert('Unknown operation type: %s', type(operation).__name__)

            actionName = '_do%s' % getCapitalizedName(operation)
            action = getattr(self.dispatch_module, actionName, None)
            if callable(action):
                return action

            ##    action = getattr(self.dispatch_module, operation, None)
            ##    if callable(action):
            ##        return action

    def dispatch(self, peer, operation, params):
        action = self.getDispatchAction(operation)
        if callable(action):
            return action(peer, *params)
        else:
            playerAlert('Unknown operation: %r' % operation)

    @classmethod
    def isInstalled(self, peer):
        pcs = getattr(peer, 'process_telnet_message', None)
        if pcs is not None:
            x = getattr(pcs, '__class__', None)
            if x and issubclass(x, self):
                return True
            x = getattr(pcs, 'im_class', None)
            if x and issubclass(x, self):
                return True

    def __init__(self, peer, dispatch_module):
        self.dispatch_module = dispatch_module
        self.install(peer)

    def install(self, peer):
        peer.process_telnet_message = self
    def uninstall(self, peer):
        try: del peer.process_telnet_message
        except AttributeError: pass

    def __call__(self, *args, **kwd):
        return self.process(*args, **kwd)
    def __repr__(self):
        return '%s (%r)' % (self.__class__.__name__, self.dispatch_module)

    __str__ = __repr__

def isProcessorInstalled(peer):
    return TelnetMessenger.isInstalled(peer)
def installProcessor(peer, *args, **kwd):
    TelnetMessenger(peer, *args, **kwd)
def uninstallProcessor(peer):
    try: pcs = peer.process_telnet_message
    except AttributeError: pass
    else:
        try: pcs.uninstall(peer)
        except AttributeError:
            del peer.process_telnet_message

# Interface.
from stuphmud.server.player import ACMD

@ACMD('teln*et-control')
def doManageTelnetControl(peer, cmd, argstr):
    args = (argstr or '').split()
    larg = len(args) and args[0].lower() or ''

    if larg in ('', 'status', 'show'):
        target = peer
        if len(args) == 2 and peer.avatar and peer.avatar.level >= 115:
            name = args[1]
            target = peer.avatar.findplayer(name, False, False)
            if not target:
                print('Player %r not found.' % name, file=peer)
            elif not target.peer:
                print('%s is not logged in.' % target.name, file=peer)
                target = None
            else:
                print('Telnet status for: %s' % target.name, file=peer)
                target = target.peer

        if target is not None:
            print('Processor is %sinstalled.' % \
                  (not isProcessorInstalled(target) and 'not ' or ''), file=peer)

            conv = getattr(target, 'telnet_conversation', None)
            if conv is not None:
                size = conv.message_size
                done = size - conv.message_left
                percent = (done / float(size)) * 100

                print('Message progress: %d/%d (%d%%)' % (done, size, percent), file=peer)
                buf = conv.message_buf
                if buf is not None:
                    print('Message sample: %r' % buf.getvalue()[:60], file=peer)

    elif larg == 'manage':
        if not isProcessorInstalled(peer):
            installProcessor(peer, getDefaultDispatchModule())
            print('Processor installed.', file=peer)
        else:
            print('Already managed.', file=peer)

    elif larg == 'cancel':
        if isProcessorInstalled(peer) >= 2 or len(args) == 2 and args[1] == 'force':
            uninstallProcessor(peer)
            print('Processor uninstalled.', file=peer)
        else:
            print('Not currently managing.', file=peer)

    elif larg == 'flush':
        delattr(peer, 'telnet_conversation')
        print('Flushed.', file=peer)

    else:
        print('Unknown subcommand: %r' % larg, file=peer)

    return True

# Dispatch Groups.
class Organized:
    def __init__(self, *modules, **kwd):
        modules = list(modules)
        self.modules = {'': modules}

        # Overload by prefixed name, retaining default.
        m = kwd.get('')
        if hasattr(m, 'extend'):
            m.extend(modules)

        self.modules.update(kwd)

    def getDispatchAction(self, name):
        # Todo: complexify the prefixed qualifier.
        i = name.find(':')
        if i >= 0:
            section = name[:i]
            name = name[i+1:]
        else:
            section = ''

        for m in self.modules.get(section, []):
            action = m.getDispatchAction(name)
            if action is not None:
                return action

    def addModule(self, module, prefix = None):
        prefix = prefix or ''
        m = self.modules.get(prefix)
        if m is None:
            m = self.modules[prefix] = []

        m.append(module)

    def __getitem__(self, name):
        return list(self.modules.get(name, []))
    def __repr__(self):
        return '%s (%s)' % (self.__class__.__name__,
                            ', '.join(list(self.modules.keys())))

    __str__ = __repr__


# Todo: write a decoupled registration api.
DEFAULT_MODULES = \
    dict(ftp = 'mud.player.telnet.ftp.FTPDispatchModule',
         rexec  = 'mud.player.telnet.rexec.RExecDispatchModule',
         # deploy = 'mud.player.telnet.deploy.LiveDeployDispatchModule',
         # http = 'mud.player.telnet.http',
         # pentacle = 'mud.player.telnet.pentacle',
         )

def _getDefaults():
    from stuphos import getSection
    section = getSection('TelnetModules')
    options = tuple((n, section.get(n)) for n in section.options())
    return options or iter(DEFAULT_MODULES.items()) # todo: merge opts with builtins?

def findDispatchModule(name):
    # A straight import will fail if the trailing members are not modules.
    # return __import__(name, globals(), locals(), fromlist = [''])

    try: module = __import__(name)
    except ImportError as e:
        e = str(e)
        if e.startswith('No module named '):
            fail = e[16:]
            if name.endswith(fail):
                major = name[:-(len(fail) + 1)]
                module = __import__(major)
            else:
                reraiseSystemException()
        else:
            reraiseSystemException()

    def getModuleMember(module, name):
        return getattr(module, name)

    parts = name.split('.')
    return reduce(getModuleMember, parts[1:], module)

def buildDefaultDispatchModule():
    dispatch = Organized()
    for (prefix, name) in _getDefaults():
        prefix = prefix.strip()
        if prefix.lower() == 'default':
            prefix = ''

        module = findDispatchModule(name)
        if module is not None:
            # Note: this is not a factory; use the class found.
            dispatch.addModule(module, prefix = prefix)

    return dispatch

from types import ClassType, MethodType, TypeType
class DispatchModuleBase:
    # Base implementation for pluggable subclasses.
    @classmethod
    def getDispatchAction(self, action):
        method = getattr(self, '_do%s' % getCapitalizedName(action), None)
        if type(method) is (ClassType, TypeType):
            # Hmm, kind of weird, but it's for modules that implement their
            # action handlers like so:
            #    class _doAction(object):
            #        def __new__(self, peer, *params):
            #            return 'result'
            #
            return method

        if callable(method):
            return MethodType(method, self())

    def _validateMessage(self, digest, key, *parts):
        return ValidateMessage(digest, key, *parts)

    def _loadKeyfile(self, keyfile):
        return LoadKeyfile(keyfile)

import hmac
def ValidateMessage(digest, key, *parts):
    hash = hmac.new(key)
    for p in parts:
        hash.update(p)

    return hash.hexdigest() == digest

def LoadKeyfile(keyfile):
    # As passphrase.
    return open(keyfile).read().strip()

# Shared COM -- todo runtime registry access.
TELNET_DISPATCH_MODULE = 'Telnet::DispatchModule::Default'
def getDefaultDispatchModule():
    from stuphos.runtime.registry import getObject
    return getObject(TELNET_DISPATCH_MODULE,
                     create = buildDefaultDispatchModule)

@runtime.api('Telnet::Dispatch::API')
class TelnetDispatchAPI:
    @classmethod
    def RegisterModuleClass(self, prefix):
        def registerAs(moduleClass):
            DEFAULT_MODULES[prefix] = '%s.%s' % \
                (moduleClass.__module__, moduleClass.__name__)

            return moduleClass
        return registerAs



# MUD Timing Tools.
from time import time as now
from time import time as getCurrentSystemTime

from datetime import datetime

def getCurrentTime():
	return datetime.fromtimestamp(now())
date = getCurrentTime

# todo: better naming
class Elapsed:
	def __init__(self):
		self.start = now()

	@property
	def elapsed(self):
		return self.measurement.elapsed

	@property
	def datetime(self):
		return datetime.fromtimestamp(self.start)

	class Measurement:
		# Point in time, relative to start.
		def __init__(self, elapsed):
			self.mark = now()
			self.e = elapsed

		@property
		def elapsed(self):
			return self.mark - self.e.start

	def measure(self):
		return self.Measurement(self)

	measurement = property(measure)

	def __str__(self):
		return '%.2f seconds' % self.elapsed



# Python Code Sandboxing Utility
# Todo: from common tools.
from compiler.transformer import Transformer
from compiler.pycodegen import ExpressionCodeGenerator
from compiler import syntax, misc
from compiler.ast import Node, Name, CallFunc, Getattr

class Sandbox(dict):
    def parse(self, expr):
        return Transformer().parseexpr(expr)

    def compile(self, expr, filename = None):
        # Parse input to AST.
        ast = self.parse(expr)

        # Scrub out any unsafe code.
        ast = self.scrubtree(ast)

        # Compile transformed tree to bytecode.
        misc.set_filename(filename or '', ast)
        syntax.check(ast)

        gen = ExpressionCodeGenerator(ast)
        return gen.getCode()

    def eval(self, expr, filename = None):
        # I don't know if this can be blocked, so we just ignore it.
        self['__builtins__'] = {}

        return eval(self.compile(expr, filename = filename),
                    self, self)

    def scrubtree(self, ast):
        # Scrub member functions, any functions not in namespace.
        def w(n):
            # Walk the syntax tree, looking for improper Names and CallFuncs.
            if isinstance(n, Node):
                if isinstance(n, Name):
                    if n.name not in self:
                        raise NameError(n.name)

                elif isinstance(n, CallFunc):
                    if not isinstance(n.node, Name):
                        raise SyntaxError('Only simple function calls allowed')

                # Todo: hide underscored attributes.
                elif isinstance(n, Getattr):
                    if n.attrname.startswith('_'):
                        raise AttributeError(n.attrname)

                # Recursive traversal.
                for c in n.getChildren():
                    w(c)

        w(ast)
        return ast

    def derive(self, **kwd):
        sub = Sandbox(self)
        sub.update(kwd)
        return sub

    def __call__(self, expr, **kwd):
        # No filename?
        return self.derive(**kwd).eval(expr)

    # Tools:
    def toLiteral(self, expr):
        r = self.eval(expr)
        if not isinstance(r, (int, float, str, list, tuple, dict)):
            raise TypeError(type(r).__name__)

        return r

    def extract(self, expr):
        def _(node):
            if isinstance(node, Node):
                if isinstance(node, (Name, CallFunc)):
                    yield node

                for c in node.getChildren():
                    for x in _(c):
                        yield x

        return list(_(self.parse(expr)))

    def examine(self, expr):
        # Show structure.
        def _(ast, indent = 0):
            tab = '  ' * indent
            if isinstance(ast, (int, float, str)):
                print('%s%r' % (tab, ast))
            else:
                print('%s%s' % (tab, ast.__class__.__name__))

            try: children = ast.getChildren()
            except AttributeError: pass
            else:
                for c in children:
                    if c is not None:
                        _(c, indent + 1)

        _(self.parse(expr))

# Exposed (safe) functions:
def safe(function):
    function._is_safe = True
    return function

def CallOpaqueMethod(object, name, *args, **kwd):
    # Perform the method-call on an object that is otherwise not allowed.
    # This is required for the submethod pattern.
    method = getattr(object, name)
    if not getattr(method, '_is_safe', False):
        raise AttributeError('Unsafe method: ' + name)

    return method(*args, **kwd)

def print_(*args):
    print(' '.join(map(str, args)))

def percentage(a, b, n = 2):
    f = '%%.%df' % min(max(0, int(n)), 8)
    return f % ((float(a) / b) * 100)


# Default sandbox (commonly-used symbols):
##    from string import join
##    s = Sandbox(pr = print_,
##                join = join,
##                percentage = percentage,
##                str = str,
##                lbrace = '{',
##                rbrace = '}',
##                call = CallOpaqueMethod,
##                xrange = xrange)



# Managed File System.
from .. import playerAlert, PlayerAlert

from errno import EEXIST

from os.path import exists, normpath, isdir
from os.path import sep as PATH_SEP, sep as root
from os.path import join as joinpath, split as splitpath
from os import makedirs
# from string import lowercase
from string import ascii_lowercase as lowercase

from stuphos import getConfig
from stuphos.runtime import Component
from stuphos.runtime.registry import setObject as registerObject
from stuphmud.server.player import HandleCommandError

PLAYER_FILE_DIR = getConfig('player-file-dir') or 'plrfiles'
PERSONAL_PLRFILE_DIR = getConfig('personal-file-dir') or 'personal'
PROGRAMS_DIR = getConfig('player-program-dir') or 'programmes'
DATABASE_DIR = getConfig('player-database-dir') or 'db'
LIBRARY_DIR = getConfig('player-library-dir') or 'library'
SECURE_DIR = getConfig('player-secure-dir') or 'secure'
KEYFILE_DIR = joinpath(SECURE_DIR, getConfig('player-keyfile-subdir') or 'keyfiles')

def splitallpath(path):
    # Just split via the PATH_SEP into individual elements, but never
    # empty ones (that is, // is treated as /).
    result = []
    while any(c != PATH_SEP for c in path): # Only works when PATH_SEP is one char.
        (head, tail) = splitpath(path)
        result.insert(0, tail)
        path = head

    return result

def getSecureFilename(filename):
    # XXX Is this platform-compatible?
    # Pretend the filename is in root, then normalize away the
    # relative parts against this path, then strip off the root.
    parts = splitallpath(normpath(joinpath(root, filename)))
    if parts:
        return joinpath(*parts)

    return ''

def getPlayername(player):
    plrname = player.name.lower()
    plrname = ''.join(c for c in plrname if c in lowercase)
    return plrname

def validatePlayername(plrname):
    plrname = plrname.lower()
    plrname = ''.join(c for c in plrname if c in lowercase)
    assert plrname
    return plrname

# What about middle-name stacks?
def getPlayerFilename(plrname, filename, folder):
    plrfile = PLAYER_FILE_DIR
    if not exists(plrfile):
        return

    plrpath = joinpath(plrfile, plrname)
    if folder:
        plrpath = joinpath(plrpath, folder)

    try: makedirs(plrpath)
    except OSError as e:
        if e.errno != EEXIST:
            raise

    if not isdir(plrpath):
        return

    return joinpath(plrpath, filename)

def getCheckedPlayerFilename(peer, filename, folder = None, dirOp = False):
    if not peer.avatar:
        playerAlert('No avatar from which to obtain player name.')

    plrname = getPlayername(peer.avatar)
    if not plrname:
        playerAlert('Could not obtain player name.')

    filename = getSecureFilename(filename)
    if not filename and not dirOp:
        playerAlert('Unable to produce secure filename.')

    plrfile = getPlayerFilename(plrname, filename, folder)
    if not plrfile:
        playerAlert('Could not obtain player filename.')

    return plrfile


# Register API.
from stuphos.runtime.registry import setObject

# API
def apply(function, *args, **kwd):
    return function(*args, **kwd)

@apply
class API:
    NAME = 'Player::DB::ManagedFS::API'
    VALID_FOLDERS = ['PLAYER_FILE_DIR', 'PERSONAL_PLRFILE_DIR',
                     'PROGRAMS_DIR', 'DATABASE_DIR', 'LIBRARY_DIR']

    getSecureFilename = staticmethod(getSecureFilename)
    getPlayername = staticmethod(getPlayername)
    getPlayerFilename = staticmethod(getPlayerFilename)
    getCheckedPlayerFilename = staticmethod(getCheckedPlayerFilename)

    def getFolder(self, kind):
        if kind in self.VALID_FOLDERS:
            return globals()[kind]

registerObject(API.NAME, API)

# todo: Web-Extended Properties?
# todo: This should be a part of mud.player.interpret, but this doesn't
#       always occur as a bridge event.  Copyover calls interpret directly.
#       also, programmes are player parts (not peer).
#
# also, since the mud.player.enterGame handler relies on programs installed via
# this managed fs, this functionality should probably be moved into the core.
#
# also, onNewConnection can't do it because player isn't there yet.
#
from shelve import open as openShelf
# import pdb; pdb.set_trace()
class PlayerFileManager(Component):
    def onPlayerActivation(self, cltr, peer, player):
        installPlayerFileStructure(peer, player)
        installPlayerPrograms(peer, player)
        if installPlayerProperties:
            # todo: disable
            installPlayerProperties(peer, player)

class PlayerFileStructure(object):
    def __init__(self, player, pathObjectClass):
        self.player = player
        self.pathObjectClass = pathObjectClass

    def getPlayername(self):
        return getPlayername(self.player)
    def __getitem__(self, name):
        return self.pathObjectClass(getPlayerFilename(self.getPlayername(), name, None))
    def __getattr__(self, name):
        try: return self[API.getFolder(self.FolderNames[name])]
        except KeyError:
            return object.__getattribute__(self, name)

    FolderNames = dict(personal = 'PERSONAL_PLRFILE_DIR',
                       programs = 'PROGRAMS_DIR',
                       database = 'DATABASE_DIR',
                       library = 'LIBRARY_DIR')

def installPlayerFileStructure(peer, player):
    try: from path import path as pathObjectClass
    except ImportError:
        try: from common.platform.path import path as pathObjectClass
        except ImportError: return

    player.plrfiles = PlayerFileStructure(player, pathObjectClass)

def installPlayerPrograms(peer, player):
    if 0:
        # File Model.
        try:
            filename = getCheckedPlayerFilename(peer, 'python.db',
                                                folder = PROGRAMS_DIR)
        except PlayerAlert as e:
            e.deliver(peer)
        except:
            HandleCommandError(peer)
        else:
            # Plop her in.
            player.programs = openShelf(filename)

    else:
        from stuphmud.server.player.notebook import Notebook
        player.programs = Notebook.Model.OpenForPlayer(player.name).shelf

# SQLObject Extended Properties.
try:
    from sqlobject import connectionForURI, SQLObject, StringCol, PickleCol
    from sqlobject import classregistry
except ImportError:
    from warnings import warn
    warn('SQLObject not installed -- Extended Properties not available!')
    installPlayerProperties = False
else:
    from os.path import abspath

    class SqlProperties(object):
        # Don't really need separate registry, because this isn't going to change much.
        try: StoredValue = classregistry.findClass('StoredValue')
        except KeyError:
            class StoredValue(SQLObject):
                name = StringCol()
                value = PickleCol() # pickleProtocol = 0) # nonstrict?

                # todo: Stored properties for all players in a single connection.
                # player = IntegerCol() # idnum

                _connection = None

        # Table Construction.
        def __init__(self, filename): # , player_idnum
            # Hmm, new db file connection for each player?
            conn = connectionForURI('sqlite:' + abspath(filename))
            self.__dict__['_%s__connection' % self.__class__.__name__] = conn

            if not SqlProperties.StoredValue.tableExists(connection = conn):
                SqlProperties.StoredValue.createTable(connection = conn)

        # Methods.
        def loadProperty(self, name):
            try: return object.__getattribute__(self, name)
            except AttributeError:
                pass

            for sv in SqlProperties.StoredValue.selectBy(connection = self.__connection,
                                                         name = name):
                return sv.value

            raise AttributeError(name)

        def storeProperty(self, name, value):
            if name.startswith('_'):
                # Do not allow storing anything else on this class.
                raise AttributeError(name)

            conn = self.__connection
            for sv in SqlProperties.StoredValue.selectBy(connection = conn, name = name):
                sv.value = value
                break
            else:
                sv = SqlProperties.StoredValue(connection = conn, name = name, value = value)

            sv._connection = conn
            sv.sync()

        def deleteProperty(self, name):
            if not name.startswith('_'):
                conn = self.__connection
                for sv in SqlProperties.StoredValue.selectBy(connection = conn, name = name):
                    sv._connection = conn
                    sv.destroySelf()
                    return

            raise AttributeError(name)

        def enumerateProperties(self):
            return (sv.name for sv in SqlProperties.StoredValue.select(connection = self.__connection))
        def numberOfProperties(self):
            return len(list(self.enumerateProperties()))
            # return SqlProperties.StoredValue.tableSize(connection = self.__connection))

        __getattr__ = loadProperty
        __setattr__ = storeProperty
        __delattr__ = deleteProperty
        __getitem__ = loadProperty
        __setitem__ = storeProperty
        __delitem__ = deleteProperty
        __iter__    = enumerateProperties
        __len__     = numberOfProperties

    def installPlayerProperties(peer, player):
        # todo: Connect properties view to existing shared database.
        try:
            filename = getCheckedPlayerFilename(peer, 'properties.sqlite',
                                                folder = DATABASE_DIR)
        except PlayerAlert as e:
            e.deliver(peer)
        except:
            HandleCommandError(peer)
        else:
            # Plop her in.
            player.properties = SqlProperties(filename)



# Telnet Dispatch Module.
from http.server import BaseHTTPRequestHandler
from io import StringIO as new_buffer

class TelnetHTTPRequestHandler(BaseHTTPRequestHandler):
    def __init__(self, request_message, peer, dispatch_module):
        self.request_message = request_message
        self.peer = peer
        self.dispatch_module = dispatch_module
        BaseHTTPRequestHandler.__init__(self, request_message, peer, dispatch_module)

    def setup(self):
        # Make request message readable by parse_request.
        self.rfile = self.request_message
        self.wfile = self.peer.response = new_buffer()

    def finish(self):
        # Do whatever it is the dispatch does with a response...
        #    A) put it on a response queue associated with peer?
        #    B) write it back to the peer via IAC?
        ##    if not self.wfile.closed:
        ##        self.wfile.flush()
        ##    self.wfile.close()
        ##    self.rfile.close()
        pass

def _doHttpRequest(peer, request_message):
    TelnetHTTPRequestHandler(request_message, peer, __import__(__name__, fromlist = ['']))
    # peer.response

def getDispatchAction(name):
    return (name == 'request') and _doHttpRequest

# Unit Testing.
from io import StringIO

class PeerHandle(StringIO):
    pass
class RequestMessage(StringIO):
    pass

REQUEST_MESSAGE = \
'''
'''

def getTelnetRequestComponents():
    peerHandle = PeerHandle()
    requestMessage = RequestMessage(REQUEST_MESSAGE)
    dispatchModule = __import__(__name__, fromlist = [''])

    return (requestMessage, peerHandle, dispatchModule)

def makeTelnetRequestHandler():
    (peerHandle, requestMessage, dispatchModule) = getTelnetRequestComponents()
    return TelnetHTTPRequestHandler(requestMessage, peerHandle, dispatchModule)

def simple_test(argv = None):
    (peerHandle, requestMessage, dispatchModule) = getTelnetRequestComponents()

    from pdb import runcall as run
    print(run(_doHttpRequest, peerHandle, requestMessage))

# A more complete server for configuring a fuller simple_test.
class THRHWrapper(BaseHTTPRequestHandler):
    # Simple _doHttpRequest, but from an actual live server.
    def __init__(self, request, client_address, server):
        BaseHTTPRequestHandler.__init__(self, request, client_address, server)
        self.telnetRequest = makeTelnetRequestHandler()
    def __repr__(self):
        return '{thrhWrapper: handling}'
    __str__ = __repr__

    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-type", 'text/html')
        self.end_headers()

        print(generatePageResponse(self), file=self.wfile)

def generatePageResponse(self):
    this = __import__(__name__)
    # this = reload(this)
    return '<html><body>%s</body></html>' % this.generateBody(self)

def generateBody(self):
    return '%s\n%s\n' % (self, dir(self))

from http.server import HTTPServer
from _thread import start_new_thread as nth

class ServerControl:
    def is_running(self):
        return bool(getattr(self, '__running', False))
    def set_running(self, state):
        setattr(self, '__running', bool(state))

    def start(self):
        self.set_running(True)
        nth(self.serve_forever, ())
        return self
    def stop(self):
        self.set_running(False)
        self.server_close()

    def serve_forever(self):
        while self.is_running():
            self.handle_request()

class TestServer(HTTPServer, ServerControl):
    PORT = 8181
    def __init__(self):
        HTTPServer.__init__(self, ('localhost', self.PORT), THRHWrapper)
    def __repr__(self):
        return '{testServer: booted}'
    __str__ = __repr__

def server_test():
    global hs
    hs = TestServer().start()
    print(hs)
    return hs

test = server_test
if __name__ == '__main__':
    test()



from twisted.internet import protocol, reactor
from twisted.application import internet, service
from twisted.protocols import amp

class MsgServer2Portal(amp.Command):
    """
    Message Server -> Portal

    """

    key = "MsgServer2Portal"
    arguments = [(b"packed_data", Compressed())]
    errors = {Exception: b"EXCEPTION"}
    response = []

class AMPMultiConnectionProtocol(amp.AMP):
    """
    AMP protocol that safely handle multiple connections to the same
    server without dropping old ones - new clients will receive
    all server returns (broadcast). Will also correctly handle
    erroneous HTTP requests on the port and return a HTTP error response.

    """

    def __init__(self, *args, **kwargs):
        """
        Initialize protocol with some things that need to be in place
        already before connecting both on portal and server.

        """
        super(AMPMultiConnectionProtocol, self).__init__(*args, **kwargs)

    def _commandReceived(self, box):
        """
        This overrides the default Twisted AMP error handling which is not
        passing enough of the traceback through to the other side. Instead we
        add a specific log of the problem on the erroring side.

        """

    def dataReceived(self, data):
        """
        Handle non-AMP messages, such as HTTP communication.
        """

    def makeConnection(self, transport):
        """
        Swallow connection log message here. Copied from original
        in the amp protocol.

        """

    def connectionMade(self):
        """
        This is called when an AMP connection is (re-)established. AMP calls it on both sides.

        """

    def connectionLost(self, reason):
        """
        We swallow connection errors here. The reason is that during a
        normal reload/shutdown there will almost always be cases where
        either the portal or server shuts down before a message has
        returned its (empty) return, triggering a connectionLost error
        that is irrelevant. If a true connection error happens, the
        portal will continuously try to reconnect, showing the problem
        that way.
        """

    def errback(self, e, info):
        """
        Error callback.
        Handles errors to avoid dropping connections on server tracebacks.

        Args:
            e (Failure): Deferred error instance.
            info (str): Error string.

        """

    def data_in(self, packed_data):
        """
        Process incoming packed data.

        Args:
            packed_data (bytes): Pickled data.
        Returns:
            unpaced_data (any): Unpickled package

        """

    def broadcast(self, command, sessid, **kwargs):
        """
        Send data across the wire to all connections.

        Args:
            command (AMP Command): A protocol send command.
            sessid (int): A unique Session id.

        Returns:
            deferred (deferred or None): A deferred with an errback.

        Notes:
            Data will be sent across the wire pickled as a tuple
            (sessid, kwargs).

        """

    def send_FunctionCall(self, modulepath, functionname, *args, **kwargs):
        """
        Access method called by either process. This will call an arbitrary
        function on the other process (On Portal if calling from Server and
        vice versa).

        Inputs:
            modulepath (str) - python path to module holding function to call
            functionname (str) - name of function in given module
            *args, **kwargs will be used as arguments/keyword args for the
                            remote function call
        Returns:
            A deferred that fires with the return value of the remote
            function call

        """

    @FunctionCall.responder
    @catch_traceback
    def receive_functioncall(self, module, function, func_args, func_kwargs):
        """
        This allows Portal- and Server-process to call an arbitrary
        function in the other process. It is intended for use by
        plugin modules.

        Args:
            module (str or module): The module containing the
                `function` to call.
            function (str): The name of the function to call in
                `module`.
            func_args (str): Pickled args tuple for use in `function` call.
            func_kwargs (str): Pickled kwargs dict for use in `function` call.

        """

class AMPServerClientProtocol(amp.AMPMultiConnectionProtocol):
    """
    This protocol describes the Server service (acting as an AMP-client)'s communication with the
    Portal (which acts as the AMP-server)

    """

    # sending AMP data

    def connectionMade(self):
        """
        Called when a new connection is established.

        """

    def send_MsgServer2Portal(self, session, **kwargs):
        """
        Access method - executed on the Server for sending data
            to Portal.

        Args:
            session (Session): Unique Session.
            kwargs (any, optiona): Extra data.

        """
        return self.data_to_portal(MsgServer2Portal, session.sessid, **kwargs)

    # @amp.MsgStatus.responder
    # def server_receive_status(self, question):
    #     return {"status": "OK"}

    # @MsgPortal2Server.responder
    # # @amp.catch_traceback
    # def server_receive_msgportal2server(self, packed_data):
    #     """
    #     Receives message arriving to server. This method is executed
    #     on the Server.

    #     Args:
    #         packed_data (str): Data to receive (a pickled tuple (sessid,kwargs))

    #     """
    #     return {}

    # @amp.AdminPortal2Server.responder
    # @amp.catch_traceback
    # def server_receive_adminportal2server(self, packed_data):
    #     """
    #     Receives admin data from the Portal (allows the portal to
    #     perform admin operations on the server). This is executed on
    #     the Server.

    #     Args:
    #         packed_data (str): Incoming, pickled data.

    #     """
    #     return {}

class AMPClientFactory(protocol.ReconnectingClientFactory):
    """
    This factory creates an instance of an AMP client connection. This handles communication from
    the be the Evennia 'Server' service to the 'Portal'. The client will try to auto-reconnect on a
    connection error.

    """

    protocol = AMPServerClientProtocol
    # Initial reconnect delay in seconds.
    initialDelay = 1
    factor = 1.5
    maxDelay = 1
    noisy = False

    def __init__(self, server):
        """
        Initializes the client factory.

        Args:
            server (server): server instance.

        """

    def startedConnecting(self, connector):
        """
        Called when starting to try to connect to the Portal AMP server.

        Args:
            connector (Connector): Twisted Connector instance representing
                this connection.

        """

    def buildProtocol(self, addr):
        """
        Creates an AMPProtocol instance when connecting to the AMP server.

        Args:
            addr (str): Connection address. Not used.

        """
        # self.resetDelay()
        self.amp_protocol = AMPServerClientProtocol()
        self.amp_protocol.factory = self
        return self.amp_protocol

    def clientConnectionLost(self, connector, reason):
        """
        Called when the AMP connection to the MUD server is lost.

        Args:
            connector (Connector): Twisted Connector instance representing
                this connection.
            reason (str): Eventual text describing why connection was lost.

        """

    def clientConnectionFailed(self, connector, reason):
        """
        Called when an AMP connection attempt to the MUD server fails.

        Args:
            connector (Connector): Twisted Connector instance representing
                this connection.
            reason (str): Eventual text describing why connection failed.

        """

factory = AMPClientFactory(None)

amp_service = internet.TCPClient(AMP_HOST, AMP_PORT, factory)
amp_service.setName("ServerAMPClient")

application = service.Application("phApp")

services = service.MultiService()
services.setServiceParent(application)

services.addService(amp_service)

def sendCommand(protocol, command, data):
    return protocol.callRemote(command, packed_data=amp.dumps(data)).addErrback(
        protocol.errback, command.key
    )

def sendAppCommand(command, data):
    return sendCommand(factory.amp_protocol, command, data)

def send_MsgServer2Portal(data):
    return sendAppCommand(MsgServer2Portal, data)


def userCreated(user, email):
    '''
    [Network]
    user-replication: stuphos.management.authlock.userCreated

    '''

    send_MsgServer2Portal(('user-created', user.username, email))



#!python
# Copyright 2011 Clint Banis.  All rights reserved.
#
# Testing.
class Peer:
    def __init__(self, input):
        self._input_q = Queue()
        self.enqueue_input(input)

    def dequeue_input_block(self, mode = 'telnet'):
        while True:
            try: block = self._input_q.get_nowait()
            except Empty:
                return

            if mode == 'telnet':
                if type(block) is not tuple:
                    continue
                if len(block) not in (2, 3):
                    continue

            elif mode == 'line':
                if (type) is not str:
                    continue

            return block

    def input_queue_for(self, mode = 'telnet'):
        while True:
            block = self.dequeue_input_block(mode = 'telnet')
            if block:
                yield block
            else:
                break

    def enqueue_input(self, input):
        for block in input:
            self._input_q.put(block)

from optparse import OptionParser
def parse_cmdln(argv = None):
    parser = OptionParser()
    parser.add_option('-d', '--debug-level', '--debug', action = 'count', default = 0)
    parser.add_option('-m', '--message')
    parser.add_option('-F', '--message-file')
    return parser.parse_args(argv)

def generate_default_message():
    msg = '\n'.join(chr(c) * 40 for c in range(ord('a'), ord('z') + 1))
    msg += '\n'
    msg *= 10
    return msg

def generate_subnegotiation_message(options):
    if options.message:
        msg = options.message
    elif options.message_file:
        msg = open(options.message_file).read()
    else:
        msg = generate_default_message()

    for data in prepare_sudata_segments(msg):
        yield data

def inspect(**ns):
    from code import InteractiveConsole as IC
    try: import readline
    except ImportError: pass
    IC(locals = ns).interact('')

def main(argv = None):
    global DEBUG_LEVEL
    (options, args) = parse_cmdln(argv)
    DEBUG_LEVEL = options.debug_level

    if options.debug_level:
        debug()

    # peer = Peer(generate_subnegotiation_message(options))
    # input = peer.input_queue_for('telnet')

    builder = SuMessageBuilder()
    for data in generate_subnegotiation_message(options):
        msg = builder.receive_sudata(data)
        if msg is not None:
            inspect(msg = msg)

if __name__ == '__main__':
    main()



# MUD Runtime.
class Facility:
    # Generic.
    # XXX Abstract! :skip:
    NAME = 'Default::Facility'
    AUTOCREATE = False # todo: metaclass
    AUTOMANAGE = False # todo: metaclass
    COMPONENT = False

    @classmethod
    def get(self, create = False):
        from .registry import getObject
        return getObject(self.NAME, create = \
                         create and self.create)

    @classmethod
    def create(self):
        if self.COMPONENT:
            from . import newComponent
            return newComponent(self)

        return self()

    @classmethod
    def CreateNew(self, andManage = False):
        cell = [False]
        def creationWrapper():
            result = self.create()
            cell[0] = True # Only signal if creation succeeded.
            return result

        # Create and track if it's new or not.
        from .registry import getObject
        getObject(self.NAME, create = creationWrapper)

        if andManage:
            self.manage()

        return cell[0]

    @classmethod
    def destroy(self):
        from .registry import delObject
        return delObject(self.NAME)

    @classmethod
    def manage(self, *args, **kwd):
        # This should address the derived manager on the derived Facility.
        return self.Manager(self, *args, **kwd)

    def getStatus(self):
        return str(self)

    # XXX Eww.. a command manager in runtime.facilities?  Should go in mud.player... :skip:
    class Manager:
        try: from stuphos.stuphlib.constants import LVL_IMPL as IMPLEMENTOR
        except ImportError as e:
            print(f'[runtime.facilities.Manager] {e}')
            IMPLEMENTOR = 115

        MINIMUM_LEVEL = IMPLEMENTOR
        SUPERUSER = False
        VERB_NAME = '*'

        def __init__(self, facility):
            self.facility = facility

            from stuphmud.server.player import ACMD
            if self.VERB_NAME and self.VERB_NAME not in ['*']:
                assignCommand = ACMD(self.VERB_NAME)
                self.doCommand = assignCommand(self.doCommand)

        def isSuperuser(self, peer):
            # If you specify SUPERUSER, you must provide this method.
            pass

        def checkAuthority(self, peer, cmd, args):
            if not self.SUPERUSER or self.isSuperuser(peer):
                if peer.avatar and peer.avatar.level >= self.MINIMUM_LEVEL:
                    return True

        def doCommand(self, peer, cmd, args):
            if self.checkAuthority(peer, cmd, args):
                largs = args and args.strip().lower() or ''
                args = args and args.split() or ()

                try:
                    if not largs or args[0] in ['show', 'status']:
                        instance = self.facility.get()
                        if instance is not None:
                            peer.page_string(instance.getStatus())
                        else:
                            print('Not installed.', file=peer)

                    elif args[0] == 'start':
                        self.doStart(peer)
                    elif args[0] in ['stop', 'destroy']:
                        self.doStop(peer)
                    elif not self.doSubCommand(peer, cmd, args):
                        print("Unknown command: '%s'" % largs, file=peer)

                except RuntimeError as e:
                    print('&r%s&N' % e, file=peer)

                return True
        # doCommand.command = 

        def doStart(self, peer):
            print(self.facility.get(create = True), file=peer)
        def doStop(self, peer):
            print(self.facility.destroy() and \
                  'Destroyed.' or 'Unknown facility.', file=peer)

        def doSubCommand(self, peer, parent, args):
            name = 'do_%s' % (args[0],)
            subcmd = getattr(self, name, None)

            if callable(subcmd):
                subcmd(peer, parent, args[1:])
                return True

            try: usage = self.usage
            except AttributeError: pass
            else: peer.page_string(usage(peer, args))

        def do_help(self, peer, cmd, args):
            from stuphos.etc import columnize, capitalize
            subcmds = [attr[3:] for attr in dir(self) if attr.startswith('do_')]
            longest = max(list(map(len, subcmds)))

            PAGE_WIDTH = 80
            title = self.VERB_NAME.replace('*', '')
            title = '-'.join(map(capitalize, title.split('-')))
            title = title + ' Subcommands'
            peer.page_string('&y%s&N\r\n&r%s&N\r\n\r\n%s' % \
                             (title, '=' * len(title),
                              columnize(subcmds, PAGE_WIDTH / longest - 1, longest + 4)))

def CreateFacility(facilityClass, andManage = False):
    # Idiom.  Todo: andManage
    facility = facilityClass.get(create = True)
    if getattr(facility, 'AUTOMANAGE', False):
        facility.manage()

    return facility

from .architecture import LookupObject
def LoadFacility(facilityFQN):
    # LoadFacility('web.adapter.session.SessionManager')
    # facilityFQN = facilityFQN.split('.')
    # module = __import__('.'.join(facilityFQN[:-1]), globals(), locals(), fromlist = [''])
    # return CreateFacility(getattr(module, facilityFQN[-1]))

    return CreateFacility(LookupObject(facilityFQN))



from .dblib import *

class RoomReader(RecordReader):
	def parseRecord(self):
		self.recordField('name',  self.tildeString())
		self.recordField('descr', self.tildeString())

		zoneNo, flags, sector = self.readLine().split()
		self.recordField('flags', ASCIIFlagDecode(flags))
		self.recordField('sector', int(sector))

		while 1:
			line = self.readLineNoEOF(self.EOFNoDollar).strip()
			c = line[0]

			if c is 'S':
				break

			if c is '$':    # detected end of records
				return  # return None-value to terminate caller

			if c is 'D':
				dir      = int(line[1:].strip())
				descr    = self.tildeString()
				keyword  = self.tildeString()

				line = self.readLineNoEOF('Expected Exit Info')

				flags, key, roomLinkNo = line.split()

				self.roomExit(dir, int(roomLinkNo), keyword, descr, int(key), ASCIIFlagDecode(flags))

			elif c is 'E':
				self.roomDescription(self.tildeString(), self.tildeString())
			elif c is '#':
				# Raise warning accordingly??
				return int(line[1:])
			else:   # Not expecting a hash mark before 'S' terminator.
				raise FormatError('Expected D/E/S: %r' % line)

		# Parse and return next hashed vnum.
		line = self.readLineNoEOF(self.EOFNoDollar)
		c = line[0]

		if c == '$':    # detected end of records
			return  # return None-value to terminate caller

		if c == '#':    # start of next record
			return int(line[1:])

		raise FormatError("Trailing room record: '%s' (%s:%d)" % (makePrintable(c), self.filename, self.lineno))

	def roomDescription(self, keyword, text):
		r = self.getCurrentRecord()
		if 'text' in r:
			r['text'][keyword] = text
		else:
			r['text'] = {keyword:text}

	def roomExit(self, dir, roomLinkNo, keyword, descr, key, flagstr):
		ex = {'room-link':roomLinkNo,'keyword':keyword,'descr':descr,'key':key,'flags':flagstr}
		r = self.getCurrentRecord()

		if 'exits' in r:
			r['exits'][dir] = ex
		else:
			r['exits'] = {dir:ex}


## SQLObject representation.
DEFAULT_ROOM_DBFL = 'lib/etc/rooms.db'

def openRoomTables(dbpath):
	from .sql import openDatabaseTables
	return openDatabaseTables \
		(dbpath, 'StuphRoomRecord', schema_package = 'sql.room')

def convert_room_row(vnum, room, columns):
	'Convert Room Record into a mapping suitable for SQLObject row representation.'
	from .sql import getStuphZoneRecordIDByZoneVnum

	row = {'vnum' : vnum}
	for n in columns:
		if n in room:
			row[n] = room[n]

	# Join with these in a separate table.
	row['extradesc']    = buffer(str(row.get('extradesc', {})))
	row['exits']        = buffer(str(row.get('exits',     {})))

	# Resolve this.
	row['stuphZoneRecordID'] = getStuphZoneRecordIDByZoneVnum(vnum / 100)

	return row

def loadStuphRoomRecords(dbpath = DEFAULT_ROOM_DBFL, populate_with = None):
	StuphRoomRecord = openRoomTables(dbpath)[0]

	if populate_with:
		def iterate_population(p):
			if type(p) is list:
				for x in p:
					for y in iterate_population(x):
						yield y
			elif type(p) is str:
				for x in RoomReader(p).parseFile().consumeRecordSet().items():
					yield x

		columns = list(StuphRoomRecord.sqlmeta.columns.keys())

		for (vnum, room) in iterate_population(populate_with):
			StuphRoomRecord(**convert_room_row(vnum, room, columns))

	return StuphRoomRecord

def test(*args):
	room_mapping = {}
	for filename in args:
		room_mapping.update(RoomReader(filename).parseFile().consumeRecordSet())
	return room_mapping

def main(opts = []):
	if '-test' in opts:
		opts.remove('-test')

		global room_mapping
		room_mapping = test(*opts)
	else:
		n = len (opts)
		i = 0

		populate_with = []
		dbfile = DEFAULT_ROOM_DBFL

		while i < n:
			a = opts[i]

			if a in ('-new', '-new-db', '-newdb'):
				dbfile = new_database_file()
			elif a in ('-db', '-dbfile', '-dbpath'):
				i += 1
				dbfile = opts[i]
			else:
				populate_with.append(a)

			i += 1

		return loadStuphRoomRecords(dbfile, populate_with = populate_with)

if __name__ == '__main__':
	from sys import argv

	if '-debug' in argv:
		from pdb import runcall
		main_app = main

		def main(*args):
			return runcall(main_app, *args)

		argv.remove('-debug')

	R = main(argv[1:])



# A stuphMUD component.
# --
# New Role:
#    Multi-user networking library which also connects the programming application.
#
#    The important thing to note about this arrangement is that the classical (or
#	 legacy) 'mud' package is now side-by-side with the network.server modules.
#

# MUD Extension Package (Core).
    # The mud package does three things: it provides a standard python package
    # for running in-game routines from python; it provides the runtime which
    # supports the event bridge (from the game server perspective); and it provides
    # a library of feature implementation that itself is designed to be extended.

# from stuphos.runtime.core import *



from stuphos.runtime import LookupObject
def open(string):
	(proto, hostpath) = string.split('://', 1)
	assert proto != '__init__'
	proto = LookupObject('stuphos.management.protocols.%s' % proto)
	return proto.open(hostpath)



# MUD Tools.

# Tool Primitives.
def registerBuiltin(object, name = None):
    if name is None:
        name = object.__name__

    __builtins__[name] = object

def asBuiltin(function):
    registerBuiltin(function)
    return function

registerBuiltin(asBuiltin)


# System Tools.
import types

from os.path import basename, join as joinpath, dirname, normpath, abspath, splitext
from time import time as getCurrentSystemTime

import linecache
from linecache import getline as getLineFromCache, clearcache as clearLineCache, checkcache as checkLineCache
from linecache import getline

try: from json import load as fromJsonFile, loads as fromJsonString, dump as toJsonFile, dumps as toJsonString
except ImportError:
    try: from simplejson import load as fromJsonFile, loads as fromJsonString, dump as toJsonFile, dumps as toJsonString
    except ImportError:
        def jsonNotAvailable(*args, **kwd):
            raise NotImplementedError('Json methods not installed!')

        fromJsonFile = fromJsonString = toJsonFile = toJsonString = jsonNotAvailable

try: from collections import OrderedDict as ordereddict
except ImportError:
    # Provide our own implementation for < 2.7
    from .collections_hack import OrderedDict as ordereddict

from _thread import start_new_thread as _nth
def nth(function, *args, **kwd):
    return _nth(function, args, kwd)

# Also found in runtime.architecture.routines
def apply(function, *args, **kwd):
    return function(*args, **kwd)

asBuiltin(apply)


# Sub-tools.
from .debugging import breakOn, traceOn, remoteBreak, remoteTrace
from .debugging import enter_debugger, debugCall, debugCall as runcall
asBuiltin(breakOn)
asBuiltin(traceOn)
asBuiltin(remoteBreak)
asBuiltin(remoteTrace)
registerBuiltin(enter_debugger, 'debugOn')

from .errors import *
from .strings import *
from . import misc
from .misc import *
from .logs import *
from .cmdln import *

registerBuiltin(log, 'logOperation')

from . import timing

try:
    setupSubmodule(vars(), '.hashlib', 'hashlib',
                   ('new', 'md5', 'sha1', 'sha224', 'sha256', 'sha384', 'sha512'))
except ImportError:
    pass # Not available.

# Pygments.
try: from .pyg_colorizer import stuphColorFormat
except (SyntaxError, ImportError) as e:
    # print e
    pyg_colorizer = False

    def stuphColorFormat(string):
        # Identity.
        return string



# BOARD Communications Manifold.
'''
communication:
    categories:
        speech::
            say, say-to?, whisper?, emote
            shout, chat*, sing*, flame*, grat*
            tell?, clan-chat*..., quest-chat*..., group-chat...,
            echo!, zecho!, gecho!, wiznet!?, page!?, send!?

            legend::
                ? = target
                t! = gc
                x = level
                ... = select

        writing::
            boards (classic)
            mail

        other::
            bug/idea/typo

output-options::
    HTML/MIME generation?

specproc assignment:
    [Room:person.services.board.comm]
    xxxx: channel.last_room_say

    [Item:person.services.board.comm]
    xxxx: channel.conversational_item
'''

# from common.runtime import Object

from stuphos import getConfig
from stuphmud.server.player import ACMD
from stuphmud.server.zones.specials.standard import Special
from stuphos.runtime.facilities import Facility
from stuphos.runtime.architecture import Synthetic

from contextlib import contextmanager
from threading import Thread
from datetime import datetime
from time import time as now
from queue import Queue

def getCurrentTimestamp():
    return datetime.fromtimestamp(now())

boardComm = runtime.BOARD.COMM
boardCommStreaming = boardComm.Streaming

class BOARDCOMM(Facility):
    NAME = 'BOARD::COMM'

    @classmethod
    def BuildORM(self):
        @boardComm.ORM
        def ORM():
            from sqlobject import SQLObject, StringCol, BoolCol
            from sqlobject import ForeignKey, IntCol, DateTimeCol
            from sqlobject.dbconnection import ConnectionHub

            # Instantiate.
            hub = boardComm.SQLHUB(ConnectionHub)

            try:
                class Channel(SQLObject):
                    class sqlmeta:
                        #table = '"stuph:comm:channel"'
                        table = 'stuph_comm_channel'
                        # idName = 'name'

                    _connection = hub

                    name = StringCol(unique = True)
                    immortal = BoolCol(default = False)
                    level = IntCol(default = 0)
                    hasTarget = BoolCol(default = False)

                class Communicae(SQLObject):
                    class sqlmeta:
                        #table = '"stuph:comm:message"'
                        table = 'stuph_comm_message'

                    _connection = hub

                    channel = ForeignKey('Channel')
                    text = StringCol()
                    sourceIdnum = IntCol()
                    targetIdnum = IntCol(default = -1)
                    timestamp = DateTimeCol()
                    room = IntCol()
                    contentType = StringCol()

                class Tag(SQLObject):
                    class sqlmeta:
                        #table = '"stuph:comm:message:tag"'
                        table = 'stuph_comm_message_tag'

                    _connection = hub

                    channel = ForeignKey('Communicae')
                    tag = StringCol()

                ##    class Conversation(SQLObject):
                ##        class sqlmeta:
                ##            table = '"stuph:comm:conversation"'
                ##            idName = 'name'
                ##
                ##        _connection = hub
                ##        name = StringCol()

                class RoomConversation(SQLObject):
                    class sqlmeta:
                        #table = '"stuph:comm:conversation:room"'
                        table = 'stuph_comm_conversation_room'
                        #idName = 'room_vnum'

                    _connection = hub
                    room_vnum = IntCol(unique = True)
                    idnum = IntCol()

            except Exception as e:
                # already in registry..
                print('%s: %s' % (e.__class__.__name__, e))
                raise e
            else:
                return Synthetic(Channel = Channel,
                                 Communicae = Communicae,
                                 Tag = Tag,
                                 RoomConversation = RoomConversation)
        return ORM

    @classmethod
    def create(self):
        # Main Facility Bootstrap.
        database = getConfig('database', 'BOARDCOMM')
        assert database

        return self(_stuphChannels_builtin,
                    self.BuildORM(), database)

    def __init__(self, channels, orm, database):
        self.orm = orm
        self.database = database

        # This call initializes db tables as well.
        (self.channels,
         self.entities) = self.registerAllChannels(channels)

        self.roomConversations = self.hub.doInTransaction \
                                 (self.openRoomConversations)

        core = runtime[runtime.Agent.System]
        if core is not None:
            core.newToolConfig(path = 'components/services/boardcomm',
                               package = self.ToolInterface(self))

    def __registry_delete__(self):
        self.EndStream()

    class ToolInterface:
        # Native
        def __init__(self, comm):
            self._comm = comm
            self._channels = dict()

        def openChannel(self, frame, name):
            chan = AdhocChannel(name)
            self._channels[name] = chan
            chan.register(self._comm)

        def makeCommunicae(self, frame, name, player, message):
            self._channels[name].dispatchMessage(player, message)

    @property
    def dbURI(self):
        # XXX
        libdir = str(io.here)
        # libdir = abspath(normpath(getcwd()))

        libdir = libdir.replace('\\', '/')

        # Cleanup windows for sqlobject paths.
        ##    if libdir[1] == ':':
        ##        libdir = libdir[0] + libdir[2:]

        db = self.database.replace \
             ('$libdir$', libdir)

        return db

    ##    @property
    ##    def dbURI(self):
    ##        from os.path import split, sep
    ##        p = self.database.replace('/', sep).split(sep)
    ##        assert p
    ##
    ##        abs = not p[0]
    ##        p = filter(None, p)
    ##        p = sep.join(p)
    ##
    ##        if abs:
    ##            p = sep + p
    ##
    ##        return 'file://' + p

    @property
    def connection(self):
        if self.database and self.database.startswith('dbcore:'):
            from stuphos.db import dbCore
            return dbCore.getConnection(self.database[7:])

        # if self.database in ['pgcore', '(pgcore)']: # pgcore:classic
        #     return runtime['PG::Core'].connection

        from sqlobject import connectionForURI
        return connectionForURI(self.dbURI)

    @property
    def hub(self):
        hub = runtime[boardComm.SQLHUB]

        # Process connection works, but only while there are no other connections on the db held open.
        hub.threadConnection = self.connection
        #hub.processConnection = self.connection
        return hub

    def initializeDB(self, hub):
        # Create tables in order (to get around relational references).
        # for (name, table) in iteritems(dictOf(self.orm)):
        for name in ['Channel', 'Communicae', 'Tag', 'RoomConversation']:
            table = getattr(self.orm, name)

            #from pdb import set_trace; set_trace()

            # Note: each table creation must happen in its own transaction,
            # because if it fails, it will corrupt all following operations
            # in that transaction.  ifNotExists doesn't appear to be obeyed.
            try: hub.doInTransaction(table.createTable, ifNotExists = True)
            except Exception as e:
                # Todo: filter relation "xxx" already exists errors from other
                # real fatal errors.
                # Todo: syslog this.
                print('%s: %s' % (name, e))

    def registerAllChannels(self, channels):
        channelMap = dict()
        for ch in channels:
            ch.register(self)
            channelMap[ch.name] = ch

        entities = dict()

        hub = self.hub # resolve once.
        self.initializeDB(hub)

        @hub.doInTransaction
        def delegate_registration():
            for ch in channels:
                entities[ch.name] = self.registerChannel \
                                    (ch.configuration)

        return (channelMap, entities)

    def registerChannel(self, boardChannelConfiguration):
        name = boardChannelConfiguration['name']

        for ch in self.orm.Channel.selectBy(name = name): # limit = 1
            return ch

        conf = dict(name = name)

        try: conf['immortal'] = bool(boardChannelConfiguration['immortal'])
        except KeyError: pass

        try: conf['level'] = int(boardChannelConfiguration['level'])
        except (KeyError, TypeError): pass

        try: conf['hasTarget'] = bool(boardChannelConfiguration['hasTarget'])
        except KeyError: pass

        return self.orm.Channel(**conf)


    # Room Conversation Management.
    def openRoomConversations(self):
        return dict((c.room_vnum, c) for c in
                    self.orm.RoomConversation.select())

    def isConversationAt(self, vnum):
        return vnum in self.roomConversations

    def startConversationAt(self, player, avatar):
        print(self.startConversationAtVNum(avatar.location.vnum, avatar.idnum), file=player)

    def startConversationAtVNum(self, vnum, idnum):
        try: existing = self.roomConversations[vnum]
        except KeyError:
            e = self.hub.doInTransaction(self.orm.RoomConversation,
                                         room_vnum = vnum,
                                         idnum = idnum)

            self.roomConversations[vnum] = e
            return 'Conversation started at #%d.' % vnum
        else:
            return 'Conversation at #%d already opened by [#%d]' % \
                  (vnum, existing.idnum)

    def showRoomConversations(self, player):
        # todo: sort
        player.page_string('\r\n'.join('#%-20.20s : %10s' % (vnum, c.idnum)
                                       for (vnum, c)
                                       in iteritems(self.roomConversations)) + '\r\n')

    def stopConversationAt(self, player, vnum):
        try: existing = self.roomConversations[vnum]
        except KeyError:
            print('There is no conversation at #%d!' % vnum, file=player)
        else:
            @self.hub.doInTransaction
            def deleteConversation():
                existing.destroySelf()

            del self.roomConversations[vnum]
            print('Stopped conversation at #%d!' % vnum, file=player)


    # Message Production.
    @classmethod
    def MakeCommunicae(self, *args):
        comm = BOARDCOMM.get()
        if comm is not None:
            return comm.makeCommunicae(*args)

    @classmethod
    def EndStream(self):
        self.Streaming.End()

    def getCommunicaeArgs(self, configuration,
                          peerHost, player, argstr):

        try: room = player.location.vnum
        except AttributeError: room = -1

        return (configuration,
                (peerHost, player.idnum,
                 room, argstr,
                 datetime.now()))

    def makeCommunicae(self, event, configuration):
        try:
            name = configuration['name']
            channel = self.channels[name] # needed?
            entity = self.entities[name]

        except KeyError: pass # todo: log, but why

        # Note: using Dispatch method instead of Enqueue.
        # Todo: pass this instance (for room conversation dict)
        else: self.Streaming.Dispatch \
              (((channel, entity),
                self.getCommunicaeArgs \
                (configuration, *event)))

    def makeMessageCommunicae(self, channelName, player, message, configuration):
        try:
            name = configuration['name']
            channel = self.channels[name] # needed?
            entity = self.entities[name]

        except KeyError: pass # todo: log, but why

        else: self.Streaming.Enqueue(self.Streaming.CommMessage(channel, entity, player, message, configuration))

    class Streaming(Thread):
        @classmethod
        def Enqueue(self, message):
            queue = boardCommStreaming.Queue(Queue)
            queue.put(message)

            for x in range(10):
                thread = boardCommStreaming.Thread(self)

                if not thread.ended: # thread.isAlive():
                    try: thread.start()
                    except RuntimeError as e:
                        pass # print e # thread already started
                    else: break

                # Try again with new thread.
                del runtime[boardCommStreaming.Thread]

            else:
                raise RuntimeError('Tried 10 times to start thread')

        @classmethod
        def End(self, wait = False):
            try: thread = runtime[boardCommStreaming.Thread]
            except KeyError: pass
            else:
                if thread is not None:
                    del runtime[boardCommStreaming.Thread]
                    (thread.join if wait else
                     thread._Thread__stop)()

        ended = False
        def run(self):
            stream_q = runtime[boardCommStreaming.Queue]
            hub = runtime[boardComm.SQLHUB]
            orm = runtime[boardComm.ORM]

            try:
                while self.isAlive():
                    msg = stream_q.get()
                    hub.doInTransaction \
                        (self.dispatchCommMessage if isinstance(msg, self.CommMessage)
                         else self.dispatchComm,
                         *((orm,) + msg))
            finally:
                self.ended = True

        @classmethod
        def Dispatch(self, message):
            # Skip putting on the queue/starting thread.
            runtime[boardComm.SQLHUB].doInTransaction \
                (self.dispatchComm,
                 *((runtime[boardComm.ORM],) + message))

        class CommMessage:
            def __init__(self, channel, entity, player, message, configuration):
                self.channel = channel
                self.entity = entity
                self.player = player
                self.message = message
                self.configuration = configuration

            @property
            def idnum(self):
                return -1

        @classmethod
        def dispatchCommMessage(self, orm, msg):
            orm.Communicae(channel = msg.entity,
                           text = msg.message,
                           sourceIdnum = msg.idnum,
                           targetIdnum = -1, # xxx gr
                           room = -1,
                           timestamp = getCurrentTimestamp(), # XXX
                           contentType = 'text/plain').sync()

        @classmethod # for Dispatch form.
        def dispatchComm(self, orm, xxx_todo_changeme, xxx_todo_changeme1):

            # Currently, only record anything within conversational
            # rooms.  Otherwise, all communication is currently recorded,
            # and that's not what we exactly want.  What we want is a
            # framework for conditional conversation recording and
            # management.  This is one of the places to customize it.
            (channelObject,
                          channelEntity) = xxx_todo_changeme
            (configuration,
                          (peerHost, idnum,
                           room, argstr,
                           timestamp)) = xxx_todo_changeme1
            if not isConversationAt(room):
                return

            # Database Insert -- todo, conditionally on conversation.
            if 0 and configuration['name'] == 'say':
                # Todo: what about whisper, emote? shout?
                if not isConversationAt(room):
                    return

            orm.Communicae(channel = channelEntity,
                           text = argstr,
                           sourceIdnum = idnum,
                           targetIdnum = -1, # xxx gr
                           room = room,
                           timestamp = timestamp,
                           contentType = 'text/plain').sync()

            # now, tags


    @classmethod
    def GetChannel(self, name):
        instance = self.get()
        assert instance is not None

        for ch in self.orm.Channel.selectBy(name = name): # limit = 1
            return ch


def isConversationAt(vnum):
    #return vnum in _conversations
    return BOARDCOMM.get().isConversationAt(vnum)


class BOARDChannel:
    def __init__(self, xxx_todo_changeme2, name,
                 immortal = False, target = False,
                 level = None, select = False):

        (verb, command) = xxx_todo_changeme2
        self.verb = verb
        self.commandName = command
        # what about subcmds? like tell/reply are the same channel
        # actually, the cmd-object is passed to the (dispatchCommand) handler,
        # so this could be used, just need to make sure to register, and handle
        # specially..

        self.name = name
        self.immortal = immortal
        self.target = target
        self.level = level
        self.select = select

    def register(self, comm):
        verb = self.verb
        if isinstance(verb, str):
            verb = [verb]

        for verb in verb:
            register = ACMD(verb)
            register(self.dispatchCommand)

        from world import command as BuiltinCommand
        self.commandObject = BuiltinCommand(self.commandName)

    @property # memorized
    def configuration(self):
        config = dict(name = self.name)

        if self.immortal is not None:
            config['immortal'] = self.immortal
        if self.level is not None:
            config['level'] = self.level

        return config

    def dispatchCommand(self, player, cmd, argstr):
        # Main Player Interface.
        if player.avatar.isPlayer:
            BOARDCOMM.MakeCommunicae((player.host,
                                      player.avatar,
                                      argstr),
                                     self.configuration)

        # return False # Let perform_command do it because
        # it will check for level, etc.

        return False

        # XXX isn't callable in game/support.
        return self.commandObject(player, argstr)


    # Special Procedures.
    class BOARDComm_RoomSayBuffer(Special):
        # RoomConversation_LastSay
        # Could probably be used as non-room specproc, too.
        COMMANDS = dict(last = 'doLastComm')

        def doLastComm(self, player, this, parsed):
            if not player.peer: # or not player.supreme:
                return False

            # todo: other privileges

            (all, few, specified) = range(3)
            last = few

            args = parsed.argstr.split()
            if not args:
                return False
            if not args[0] == 'say':
                return False

            del args[0]
            # todo: last "here"?
            if args:
                assert len(args) == 1
                a = args[0]
                if a == 'all':
                    last = all
                elif a == 'few':
                    last = few
                elif a.isdigit():
                    limit = int(a)
                    last = specified
                else:
                    raise ValueError(a)

            if last == all:
                limit = None
            elif last == few:
                limit = 15

            vnum = player.location.vnum

            # Instance of BOARDCOMM
            instance = BOARDCOMM.get()
            assert instance is not None

            for ch in instance.orm.Channel.selectBy(name = 'say'): # limit = 1
                break
            else:
                raise NameError(name)

            def makeQuery():
                for c in instance.orm.Communicae.selectBy \
                    (channel = ch, room = vnum).limit(limit) \
                    .orderBy('-timestamp'):
                    #(channel = ch, room = vnum).limit(limit):
                    # todo: order by so that 'last' has any meaning...
                    # .orderBy('timestamp'):
                    yield c

            def reportQuery(g):
                from world.player import playerByID

                for (i, c) in enumerate(reversed(list(g))): # XXX reversing here
                    # channel, text, sourceIdnum, targetIdnum, room, timestamp
                    idnum = c.sourceIdnum

                    # todo: cache these lookups.
                    try: name = playerByID(idnum)
                    except ValueError: name = '#%d' % idnum

                    yield '[%s] (%-20s) %s' % (c.timestamp, name, c.text)

                yield ''

            player.peer.page_string('\r\n'.join(reportQuery(makeQuery())))


    class BOARDComm_SpecialInterface_Buffer(Special):
        # Not sure what to do with this:
        ##    def StuphOS_BOARD_Pack(self, item):
        ##        pass

        TARGET_SELF = True
        COMMANDS = dict(read = 'doReadComm',
                        look = 'doReadComm')

        def __init__(self, **values):
            self.name = values['channel-name']
            self.query = values['query']

        def doReadComm(self, player, this, parsed):
            args = parsed.argstr.split()
            if args:
                #@Threading
                def performQuery():
                    # Create a report object, put it in player's inventory,
                    # or associate it with the buffer as a BOARD message,
                    # and then notify.
                    with BOARDCOMM.GetChannel(self.name) as channel:
                        return channel.entity.select(**self.query)

                builtin(pq = performQuery)
                return True

    class BOARDComm_SpecialInterface_ConversationManagement:
        class Base(Special):
            # Manage conversational contexts, in generic.
            COMMANDS = {':': 'doManageConversations'}

            def doManageConversations(self, player, this, parsed):
                #:conversation start here

                if not player.peer or not player.supreme:
                    return False

                args = self.prepareArgs(this, player, parsed)

                if args[0] == 'start':
                    del args[0]
                    assert args and args[0] == 'here'

                    BOARDCOMM.get().startConversationAt(player.peer, player)

                elif args[0] in ['show', 'list']:
                    del args[0]
                    assert not args

                    BOARDCOMM.get().showRoomConversations(player.peer)

                elif args[0] in ['stop']:
                    del args[0]
                    assert len(args) == 1

                    if args[0] == 'here':
                        location = player.location.vnum
                    else:
                        assert args[0].isdigit()
                        location = int(args[0])

                    BOARDCOMM.get().stopConversationAt(player.peer, location)

                else:
                    # Unknown
                    return False

                return True


        # Specific interfaces:
        class Room(Base):
            def prepareArgs(self, this, player, parsed):
                args = parsed.argstr.split()
                assert args
                assert args[0] == 'conversations'
                del args[0]

                return args

            ##    AttributeError: 'room' object has no attribute 'page_string'
            ##                                 [ixion\stuph\lib\python\mud\zones\specials\standard.py:184] __call__
            ##                                   this, parsed) is False:
            ##                                 [ixion\stuph\lib\python\mud\zones\specials\standard.py:197] dispatchCommand
            ##                                   return action(*args, **kwd)
            ##                                 [ixion\stuph\lib\python\packages\implementors\person\services\board\comm.py:557] doManageConversations
            ##                                   BOARDCOMM.get().showRoomConversations(player)
            ##                                 [ixion\stuph\lib\python\packages\implementors\person\services\board\comm.py:257] showRoomConversations
            ##                                   player.page_string('\r\n'.join('#%-20.20s : %10s' % (vnum, c.idnum)
            ##    Unknown command: ':conversations list'

        class Item(Base):
            TARGET_SELF = True

            def prepareArgs(self, this, player, parsed):
                args = parsed.argstr.split()
                del args[0] # name of object
                assert args

                return args

class AdhocChannel(BOARDChannel):
    def __init__(self, name, instance = None):
        self.name = name
        self.instance = instance

    @property # memorized
    def configuration(self):
        return dict()

    def register(self, comm):
        pass

    def dispatchMessage(self, player, message):
        comm = BOARDCOMM.get()
        if comm is not None:
            comm.makeMessageCommunicae(self.name, player, message, configuration)

_stuphChannels_builtin = [BOARDChannel((['say*', "'"      ], 'say'), 'say'      ),
                          BOARDChannel((['to*', '>'       ], 'to'), 'sayTo',
                                       target = True),

                          BOARDChannel((['whisper*'       ], 'whisper'), 'whisper'  ),
                          BOARDChannel((['emote*', ':'    ], 'emote'), 'emote',
                                       target = True),

                          BOARDChannel((['chat*', '.'     ], 'chat'), 'chat'     ),
                          BOARDChannel((['shout*'         ], 'shout'), 'shout'    ),
                          BOARDChannel((['sing*'          ], 'sing'), 'sing'     ),
                          BOARDChannel((['flame*'         ], 'flame'), 'flame'    ),
                          BOARDChannel((['grat*'          ], 'grats'), 'congrats'   ),

                          BOARDChannel((['tell*'          ], 'tell'), 'tell',
                                       target = True),
                          # reply?

                          BOARDChannel((['csay*', 'ctell*'], 'csay'), 'clanChat',
                                       select = True),

                          BOARDChannel((['qsay*'          ], 'qsay'), 'questChat',
                                       select = True),

                          BOARDChannel((['gsay*'          ], 'gsay'), 'groupChat',
                                       select = True),

                          BOARDChannel((['echo*'          ], 'echo'), 'echo',
                                       target = False, immortal = True),

                          BOARDChannel((['zecho*'         ], 'zecho'), 'zecho',
                                       target = False, immortal = True),

                          BOARDChannel((['gecho*'         ], 'gecho'), 'gecho',
                                       target = False, immortal = True),

                          # Todo: remove '-' from channeling for sake of wizcontrol??
                          BOARDChannel((['wiznet*', '-'   ], 'wiznet'), 'wiznet',
                                       target = False, immortal = True),

                          BOARDChannel((['send*'          ], 'send'), 'send',
                                       target = True, immortal = True),

                          BOARDChannel((['page*'          ], 'page'), 'page',
                                       target = True, immortal = True)]


# Query and report from within main module during testing:
# for (_, _, msg, _, _, time, _) in sqlite.Connection(c.connection).sqlite3Connection.execute('select * from "stuph:comm:message"'): print msg

##    if 0:
##        stuphMud.testing.module.BOARDChannel.BOARDComm_SpecialInterface_Buffer \
##            (**{'query': '', 'channel-name': ''}) \
##            (core.console, None, synthetic(name = 'read'), 'query')
##
##        o = stuphMud.testing.module.BOARDChannel.BOARDComm_RoomSayBuffer()
##        o(None, core.console.avatar, synthetic(name = 'last'), '')
##
##        RoomConversation_LastSay = BOARDComm_RoomSayBuffer

# Specproc Assignment Interface.
class channel:
    last_room_say = BOARDChannel.BOARDComm_RoomSayBuffer
    last_say = last_room_say
    conversational_room = BOARDChannel.BOARDComm_SpecialInterface_ConversationManagement.Room
    conversational_item = BOARDChannel.BOARDComm_SpecialInterface_ConversationManagement.Item

# selectBy(room = room).limit(10)


# _conversations = dict()

# runtime[boardComm.RoomConversations]

##    @ACMD('boardcomm-*admin')
##    def doAdminBOARDCOMM(player, cmd, argstr):
##        avatar = player.avatar
##        if avatar and avatar.supreme:
##            args = argstr.split()
##            if args:
##                if args[0] == 'conversation':
##                    del args[0]
##                    assert args
##
##                    if args[0] == 'start':
##                        del args[0]
##                        assert args and args[0] == 'here'
##
##                        BOARDCOMM.get().startConversationAt(player, avatar)
##                        ##    vnum = avatar.location.vnum
##                        ##
##                        ##    try: existing = _conversations[vnum]
##                        ##    except KeyError:
##                        ##        _conversations[vnum] = dict(idnum = avatar.idnum)
##                        ##        print >> player, 'Conversation started at #%d.' % vnum
##                        ##    else:
##                        ##        print >> player, 'Conversation at #%d already opened by [#%d]' % \
##                        ##              (vnum, existing['idnum'])
##
##                    elif args[0] in ['show', 'list']:
##                        del args[0]
##                        assert not args
##
##                        BOARDCOMM.get().showRoomConversations(player)
##                        ##    player.page_string('\r\n'.join('#%-20.20s : %10s' % nv for nv
##                        ##                                   in iteritems(_conversations)))
##
##                    elif args[0] in ['stop']:
##                        assert not args # for now
##
##                    else:
##                        return False
##
##                    return True


# Generalized:
class ConversationManagement:
    def handleCommand(self, player, this, args, parsed):
        if args[0] == 'start':
            del args[0]
            assert args and args[0] == 'here'

            BOARDCOMM.get().startConversationAt(player, player.avatar)
            ##    vnum = avatar.location.vnum
            ##
            ##    try: existing = _conversations[vnum]
            ##    except KeyError:
            ##        _conversations[vnum] = dict(idnum = avatar.idnum)
            ##        print >> player, 'Conversation started at #%d.' % vnum
            ##    else:
            ##        print >> player, 'Conversation at #%d already opened by [#%d]' % \
            ##              (vnum, existing['idnum'])

        elif args[0] in ['show', 'list']:
            del args[0]
            assert not args

            BOARDCOMM.get().showRoomConversations(player)
            ##    player.page_string('\r\n'.join('#%-20.20s : %10s' % nv for nv
            ##                                   in iteritems(_conversations)))

        elif args[0] in ['stop']:
            assert not args # for now

        else:
            return False

        return True

##    @ACMD('boardcomm-*admin')
##    @apply
##    class WizardConversationMngr(ConversationManagement):
##        def __call__(self, player, cmd, argstr):
##            avatar = player.avatar
##            if avatar and avatar.supreme:
##                args = argstr and argstr.split() or []
##                if args:
##                    if args[0] == 'conversation':
##                        del args[0]
##                        assert args
##
##                        return self.handleCommand(player, None, args, None)



# Compartment Server.
# (C) 2021 runphase.com .  All rights reserved.
# --
# This comprises the standalone server provision that hosts the application,
# by building on stuphos.system, and also contains an implementation
# of the game module, such that the embedding aspects of the application
# software can perform where the extension protocol is expected.
#
#     So this means that embedding/extension aspects that are throughout
#     the application package need to be unified in some independent
#     package, but because they are still integrated, this compartment
#     is built from the synthesis of a game module for a non-embedding
#     provision that implements extension protocol.
#
#     The nominal 'game' module is provided parallel to this package,
#     adopts all of the functionality of this package for protocol
#     purposes, but this is the package that you run when not embedding.
#

import sys
import warnings

warnings.warn("module '%s' is not native" % __name__,
              ImportWarning, stacklevel = 1)

version = '1.0'

framework = True
embedded = False

import world
sys.modules['world'] = world

# Older versions of pickles might rely on this existing.
sys.modules['game.world'] = world

syslog_cr = True # Add CR to end of syslog lines.


def installWorldBridge():
    # Provide role of game bindings.
    world.installEventBridge()

def bridgeModule():
    # This is required by the mud runtime.
    return world

def syslog(*args):
    args = ' '.join(map(str, args))
    if syslog_cr:
        args += '\r'

    # debugOn()
    print(args, file = sys.stderr)

# Construct engine operations -- Do this after defining the above bit of game.
# XXX This should be done outside of the gamemodule (in the Extension Library) :skip:

# Yeah, because this is a gamemodule replacement.  So this doesn't have any
# purpose except where wrapping the dll..

# Currently causes segv (stackdump) crash on exit:
##    try: from mud.api import Main
##    except ImportError: pass
##    else: mainOps = Main()

def installBridge(bridge = None):
    # Install Event Bridge onto MUD-Package object.
    # This works sensibly because bootStart isn't on the bridge.
    if bridge is None:
        import builtins as builtin

        try: bridge = builtin.StuphMUD
        except AttributeError:
            from stuphos.runtime import EventBridge
            bridge = builtin.StuphMUD = EventBridge()

    for (api, event) in (('NewIncomingPeer'    , 'newConnection' ),
                         ('GreetPlayer'        , 'greetPlayer'   ),
                         ('FirstSpecialCommand', 'firstSpecial'  ),
                         ('LastSpecialCommand' , 'lastSpecial'   ),
                         ('StartWorldReset'    , 'resetStart'    ),
                         ('CompleteWorldReset' , 'resetComplete')):

        setattr(bridge, api, event)

    installWorldBridge()
